import io
import sys


class Socket:
    def __init__(self):
        self._write_buffer = io.BytesIO()
        self._read_buffer = io.BytesIO(b"HTTP/1.0 200 OK\r\n\r\n")

    def connect(self, address):
        pass

    def write(self, buf):
        self._write_buffer.write(buf)

    def readline(self):
        return self._read_buffer.readline()


class socket:
    AF_INET = 2
    SOCK_STREAM = 1
    IPPROTO_TCP = 6

    @staticmethod
    def getaddrinfo(host, port, af=0, type=0, flags=0):
        return [(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP, "", ("127.0.0.1", 80))]

    def socket(af=AF_INET, type=SOCK_STREAM, proto=IPPROTO_TCP):
        return Socket()


sys.modules["socket"] = socket
# ruff: noqa: E402
import requests


def format_message(response):
    return response.raw._write_buffer.getvalue().decode("utf8")


def test_simple_get():
    response = requests.request("GET", "http://example.com")

    assert response.raw._write_buffer.getvalue() == (
        b"GET / HTTP/1.0\r\n" + b"Connection: close\r\n" + b"Host: example.com\r\n\r\n"
    ), format_message(response)


def test_get_auth():
    response = requests.request(
        "GET", "http://example.com", auth=("test-username", "test-password")
    )

    assert response.raw._write_buffer.getvalue() == (
        b"GET / HTTP/1.0\r\n"
        + b"Host: example.com\r\n"
        + b"Authorization: Basic dGVzdC11c2VybmFtZTp0ZXN0LXBhc3N3b3Jk\r\n"
        + b"Connection: close\r\n\r\n"
    ), format_message(response)


def test_get_custom_header():
    response = requests.request("GET", "http://example.com", headers={"User-Agent": "test-agent"})

    assert response.raw._write_buffer.getvalue() == (
        b"GET / HTTP/1.0\r\n"
        + b"User-Agent: test-agent\r\n"
        + b"Host: example.com\r\n"
        + b"Connection: close\r\n\r\n"
    ), format_message(response)


def test_post_json():
    response = requests.request("GET", "http://example.com", json="test")

    assert response.raw._write_buffer.getvalue() == (
        b"GET / HTTP/1.0\r\n"
        + b"Connection: close\r\n"
        + b"Content-Type: application/json\r\n"
        + b"Host: example.com\r\n"
        + b"Content-Length: 6\r\n\r\n"
        + b'"test"'
    ), format_message(response)


def test_post_chunked_data():
    def chunks():
        yield "test"

    response = requests.request("GET", "http://example.com", data=chunks())

    assert response.raw._write_buffer.getvalue() == (
        b"GET / HTTP/1.0\r\n"
        + b"Transfer-Encoding: chunked\r\n"
        + b"Host: example.com\r\n"
        + b"Connection: close\r\n\r\n"
        + b"4\r\ntest\r\n"
        + b"0\r\n\r\n"
    ), format_message(response)


def test_overwrite_get_headers():
    response = requests.request(
        "GET", "http://example.com", headers={"Host": "test.com", "Connection": "keep-alive"}
    )

    assert response.raw._write_buffer.getvalue() == (
        b"GET / HTTP/1.0\r\n" + b"Connection: keep-alive\r\n" + b"Host: test.com\r\n\r\n"
    ), format_message(response)


def test_overwrite_post_json_headers():
    response = requests.request(
        "GET",
        "http://example.com",
        json="test",
        headers={"Content-Type": "text/plain", "Content-Length": "10"},
    )

    assert response.raw._write_buffer.getvalue() == (
        b"GET / HTTP/1.0\r\n"
        + b"Connection: close\r\n"
        + b"Content-Length: 10\r\n"
        + b"Content-Type: text/plain\r\n"
        + b"Host: example.com\r\n\r\n"
        + b'"test"'
    ), format_message(response)


def test_overwrite_post_chunked_data_headers():
    def chunks():
        yield "test"

    response = requests.request(
        "GET", "http://example.com", data=chunks(), headers={"Content-Length": "4"}
    )

    assert response.raw._write_buffer.getvalue() == (
        b"GET / HTTP/1.0\r\n"
        + b"Host: example.com\r\n"
        + b"Content-Length: 4\r\n"
        + b"Connection: close\r\n\r\n"
        + b"test"
    ), format_message(response)


def test_do_not_modify_headers_argument():
    global do_not_modify_this_dict
    do_not_modify_this_dict = {}
    requests.request("GET", "http://example.com", headers=do_not_modify_this_dict)

    assert do_not_modify_this_dict == {}, do_not_modify_this_dict


test_simple_get()
test_get_auth()
test_get_custom_header()
test_post_json()
test_post_chunked_data()
test_overwrite_get_headers()
test_overwrite_post_json_headers()
test_overwrite_post_chunked_data_headers()
test_do_not_modify_headers_argument()



d = {1:2, 3:4}
print(len(d))
d.update(["ab"])
print(d[1])
print(d[3])
print(d["a"])
print(len(d))
d.update([(1,4)])
print(d[1])
print(len(d))

# using keywords
d.update(a=5)
print(d['a'])
d.update([(1,5)], b=6)
print(d[1], d['b'])



# Test construction of machine.UART objects.

import sys
from machine import UART

print(UART(0, tx=0, rx=1))

if "RP2350" in sys.implementation._machine:
    # Test that UART can be constructed using other tx/rx pins.
    UART(0, tx=2, rx=3)



import sys

# ruff: noqa: E402
sys.path.insert(0, ".")
import aiohttp
import asyncio

headers = {"Accept-Encoding": "gzip,deflate"}


async def main():
    async with aiohttp.ClientSession(headers=headers, version=aiohttp.HttpVersion11) as session:
        async with session.get("http://micropython.org") as response:
            print("Status:", response.status)
            print("Content-Type:", response.headers["Content-Type"])
            print(response.headers)
            html = await response.text()
            print(html)


asyncio.run(main())



import time
import binascii
import machine
from umqtt.simple import MQTTClient
from machine import Pin


# Many ESP8266 boards have active-low "flash" button on GPIO0.
button = Pin(0, Pin.IN)

# Default MQTT server to connect to
SERVER = "192.168.1.35"
CLIENT_ID = binascii.hexlify(machine.unique_id())
TOPIC = b"led"


def main(server=SERVER):
    c = MQTTClient(CLIENT_ID, server)
    c.connect()
    print("Connected to %s, waiting for button presses" % server)
    while True:
        while True:
            if button.value() == 0:
                break
            time.sleep_ms(20)
        print("Button pressed")
        c.publish(TOPIC, b"toggle")
        time.sleep_ms(200)

    c.disconnect()



from utime import *
from micropython import const

_TS_YEAR = const(0)
_TS_MON = const(1)
_TS_MDAY = const(2)
_TS_HOUR = const(3)
_TS_MIN = const(4)
_TS_SEC = const(5)
_TS_WDAY = const(6)
_TS_YDAY = const(7)
_TS_ISDST = const(8)

_WDAY = const(("Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"))
_MDAY = const(
    (
        "January",
        "February",
        "March",
        "April",
        "May",
        "June",
        "July",
        "August",
        "September",
        "October",
        "November",
        "December",
    )
)


def strftime(datefmt, ts):
    from io import StringIO

    fmtsp = False
    ftime = StringIO()
    for k in datefmt:
        if fmtsp:
            if k == "a":
                ftime.write(_WDAY[ts[_TS_WDAY]][0:3])
            elif k == "A":
                ftime.write(_WDAY[ts[_TS_WDAY]])
            elif k == "b":
                ftime.write(_MDAY[ts[_TS_MON] - 1][0:3])
            elif k == "B":
                ftime.write(_MDAY[ts[_TS_MON] - 1])
            elif k == "d":
                ftime.write("%02d" % ts[_TS_MDAY])
            elif k == "H":
                ftime.write("%02d" % ts[_TS_HOUR])
            elif k == "I":
                ftime.write("%02d" % (ts[_TS_HOUR] % 12))
            elif k == "j":
                ftime.write("%03d" % ts[_TS_YDAY])
            elif k == "m":
                ftime.write("%02d" % ts[_TS_MON])
            elif k == "M":
                ftime.write("%02d" % ts[_TS_MIN])
            elif k == "P":
                ftime.write("AM" if ts[_TS_HOUR] < 12 else "PM")
            elif k == "S":
                ftime.write("%02d" % ts[_TS_SEC])
            elif k == "w":
                ftime.write(str(ts[_TS_WDAY]))
            elif k == "y":
                ftime.write("%02d" % (ts[_TS_YEAR] % 100))
            elif k == "Y":
                ftime.write(str(ts[_TS_YEAR]))
            else:
                ftime.write(k)
            fmtsp = False
        elif k == "%":
            fmtsp = True
        else:
            ftime.write(k)
    val = ftime.getvalue()
    ftime.close()
    return val



import sys

# ruff: noqa: E402
sys.path.append("")

from micropython import const

import asyncio
import aioble
import bluetooth

import random
import struct

# org.bluetooth.service.environmental_sensing
_ENV_SENSE_UUID = bluetooth.UUID(0x181A)
# org.bluetooth.characteristic.temperature
_ENV_SENSE_TEMP_UUID = bluetooth.UUID(0x2A6E)
# org.bluetooth.characteristic.gap.appearance.xml
_ADV_APPEARANCE_GENERIC_THERMOMETER = const(768)

# How frequently to send advertising beacons.
_ADV_INTERVAL_MS = 250_000


# Register GATT server.
temp_service = aioble.Service(_ENV_SENSE_UUID)
temp_characteristic = aioble.Characteristic(
    temp_service, _ENV_SENSE_TEMP_UUID, read=True, notify=True
)
aioble.register_services(temp_service)


# Helper to encode the temperature characteristic encoding (sint16, hundredths of a degree).
def _encode_temperature(temp_deg_c):
    return struct.pack("<h", int(temp_deg_c * 100))


# This would be periodically polling a hardware sensor.
async def sensor_task():
    t = 24.5
    while True:
        temp_characteristic.write(_encode_temperature(t), send_update=True)
        t += random.uniform(-0.5, 0.5)
        await asyncio.sleep_ms(1000)


# Serially wait for connections. Don't advertise while a central is
# connected.
async def peripheral_task():
    while True:
        async with await aioble.advertise(
            _ADV_INTERVAL_MS,
            name="mpy-temp",
            services=[_ENV_SENSE_UUID],
            appearance=_ADV_APPEARANCE_GENERIC_THERMOMETER,
        ) as connection:
            print("Connection from", connection.device)
            await connection.disconnected(timeout_ms=None)


# Run both tasks.
async def main():
    t1 = asyncio.create_task(sensor_task())
    t2 = asyncio.create_task(peripheral_task())
    await asyncio.gather(t1, t2)


asyncio.run(main())



# LSM6DSOX Basic Example.
import time
from lsm6dsox import LSM6DSOX

from machine import Pin, I2C

lsm = LSM6DSOX(I2C(0, scl=Pin(13), sda=Pin(12)))
# Or init in SPI mode.
# lsm = LSM6DSOX(SPI(5), cs=Pin(10))

while True:
    print("Accelerometer: x:{:>8.3f} y:{:>8.3f} z:{:>8.3f}".format(*lsm.accel()))
    print("Gyroscope:     x:{:>8.3f} y:{:>8.3f} z:{:>8.3f}".format(*lsm.gyro()))
    print("")
    time.sleep_ms(100)



import sys

# ruff: noqa: E402
sys.path.append("")

from micropython import const

import asyncio
import aioble
import bluetooth

import random
import struct

# org.bluetooth.service.environmental_sensing
_ENV_SENSE_UUID = bluetooth.UUID(0x181A)
# org.bluetooth.characteristic.temperature
_ENV_SENSE_TEMP_UUID = bluetooth.UUID(0x2A6E)


# Helper to decode the temperature characteristic encoding (sint16, hundredths of a degree).
def _decode_temperature(data):
    return struct.unpack("<h", data)[0] / 100


async def find_temp_sensor():
    # Scan for 5 seconds, in active mode, with very low interval/window (to
    # maximise detection rate).
    async with aioble.scan(5000, interval_us=30000, window_us=30000, active=True) as scanner:
        async for result in scanner:
            # See if it matches our name and the environmental sensing service.
            if result.name() == "mpy-temp" and _ENV_SENSE_UUID in result.services():
                return result.device
    return None


async def main():
    device = await find_temp_sensor()
    if not device:
        print("Temperature sensor not found")
        return

    try:
        print("Connecting to", device)
        connection = await device.connect()
    except asyncio.TimeoutError:
        print("Timeout during connection")
        return

    async with connection:
        try:
            temp_service = await connection.service(_ENV_SENSE_UUID)
            temp_characteristic = await temp_service.characteristic(_ENV_SENSE_TEMP_UUID)
        except asyncio.TimeoutError:
            print("Timeout discovering services/characteristics")
            return

        while connection.is_connected():
            temp_deg_c = _decode_temperature(await temp_characteristic.read())
            print("Temperature: {:.2f}".format(temp_deg_c))
            await asyncio.sleep_ms(1000)


asyncio.run(main())



#!/usr/bin/env python3
"""
Code analysis tools for the NEAR Python contract compiler.
"""

import ast
import sys
import tomllib
from pathlib import Path
from typing import List, Set

from .utils import console

# MicroPython module lists for dependency analysis
MPY_MODULES = {"array", "builtins", "json", "os", "random", "struct", "sys"}
MPY_LIB_PACKAGES = {"aiohttp", "cbor2", "iperf3", "pyjwt", "requests"}
MPY_STDLIB_PACKAGES = [
    "binascii",
    "contextlib",
    "fnmatch",
    "hashlib-sha224",
    "hmac",
    "keyword",
    "os-path",
    "pprint",
    "stat",
    "tempfile",
    "types",
    "warnings",
    "__future__",
    "bisect",
    "copy",
    "functools",
    "hashlib-sha256",
    "html",
    "locale",
    "pathlib",
    "quopri",
    "string",
    "textwrap",
    "unittest",
    "zlib",
    "abc",
    "cmd",
    "curses.ascii",
    "gzip",
    "hashlib-sha384",
    "inspect",
    "logging",
    "pickle",
    "random",
    "struct",
    "threading",
    "unittest-discover",
    "argparse",
    "collections",
    "datetime",
    "hashlib",
    "hashlib-sha512",
    "io",
    "operator",
    "pkg_resources",
    "shutil",
    "tarfile",
    "time",
    "uu",
    "base64",
    "collections-defaultdict",
    "errno",
    "hashlib-core",
    "heapq",
    "itertools",
    "os",
    "pkgutil",
    "ssl",
    "tarfile-write",
    "traceback",
    "venv",
]
NEAR_MODULE_NAME = "near"


def validate_export_names(exports: Set[str]) -> List[str]:
    """
    Validate export names against C reserved keywords.

    Args:
        exports: Set of function names that are marked as NEAR exports

    Returns:
        List of invalid export names that are C keywords
    """
    c_keywords = {
        "auto",
        "break",
        "case",
        "char",
        "const",
        "continue",
        "default",
        "do",
        "double",
        "else",
        "enum",
        "extern",
        "float",
        "for",
        "goto",
        "if",
        "inline",
        "int",
        "long",
        "register",
        "restrict",
        "return",
        "short",
        "signed",
        "sizeof",
        "static",
        "struct",
        "switch",
        "typedef",
        "union",
        "unsigned",
        "void",
        "volatile",
        "while",
        # C99 and later keywords
        "_Alignas",
        "_Alignof",
        "_Atomic",
        "_Bool",
        "_Complex",
        "_Generic",
        "_Imaginary",
        "_Noreturn",
        "_Static_assert",
        "_Thread_local",
    }

    return [export for export in exports if export in c_keywords]


def find_exports(file_path: Path) -> Set[str]:
    """
    Find all functions decorated with NEAR export decorators in a Python file.

    Args:
        file_path: Path to the Python file

    Returns:
        Set of function names that are marked as NEAR exports
    """
    with open(file_path) as f:
        tree = ast.parse(f.read())

    export_decorators = {
        "export",
        "view",
        "call",
        "init",
        "callback",
        "multi_callback",
        "near.export",
    }
    exports = set()

    for node in ast.walk(tree):
        if not isinstance(node, ast.FunctionDef) or not node.decorator_list:
            continue

        for decorator in node.decorator_list:
            name = None

            # Simple name: @export
            if isinstance(decorator, ast.Name):
                name = decorator.id
            # Call: @export()
            elif isinstance(decorator, ast.Call) and isinstance(
                decorator.func, ast.Name
            ):
                name = decorator.func.id
            # Attribute: @near.export
            elif isinstance(decorator, ast.Attribute) and isinstance(
                decorator.value, ast.Name
            ):
                if decorator.value.id == "near" and decorator.attr == "export":
                    name = "near.export"

            if name in export_decorators:
                exports.add(node.name)
                break

    # Always include contract_source_metadata in exports
    # This ensures it's properly registered even if we need to inject it
    exports.add("contract_source_metadata")
    exports.add("__contract_abi")

    return exports


def find_imports(file_path: Path) -> Set[str]:
    """
    Find all imported modules in a Python file.

    Args:
        file_path: Path to the Python file

    Returns:
        Set of module names that are imported
    """
    with open(file_path) as f:
        tree = ast.parse(f.read())

    imports = set()

    for node in ast.walk(tree):
        # Direct imports: import foo, bar
        if isinstance(node, ast.Import):
            for name in node.names:
                imports.add(name.name)
        # From imports: from foo import bar
        elif isinstance(node, ast.ImportFrom) and node.module:
            imports.add(node.module)

    return imports


def is_micropython_module(module_name: str) -> bool:
    """
    Check if a module is included in MicroPython.

    Args:
        module_name: Name of the module to check

    Returns:
        True if the module is included in MicroPython, False otherwise
    """
    base_module = module_name.split(".")[0]
    return (
        base_module in MPY_MODULES
        or base_module in MPY_LIB_PACKAGES
        or base_module in MPY_STDLIB_PACKAGES
        or base_module == NEAR_MODULE_NAME
    )


def get_excluded_stdlib_packages(project_path: Path) -> List[str]:
    """
    Get excluded stdlib packages from pyproject.toml.

    Args:
        project_path: Path to the project directory

    Returns:
        List of excluded package names
    """
    pyproject_path = project_path / "pyproject.toml"
    excluded_packages = []

    if pyproject_path.is_file():
        try:
            with open(pyproject_path, "rb") as file:
                pyproject_data = tomllib.load(file)
            excluded_packages = (
                pyproject_data.get("tool", {})
                .get("near-py-tool", {})
                .get("exclude-micropython-stdlib-packages", [])
            )
        except (ImportError, Exception) as e:
            console.print(
                f"[yellow]Warning: Could not read exclusions from pyproject.toml: {e}"
            )

    return excluded_packages


def analyze_contract(contract_path: Path) -> tuple[Set[str], Set[str]]:
    """
    Analyze a contract file to find exports and imports.

    Args:
        contract_path: Path to the contract file

    Returns:
        Tuple of (exports, imports)
    """
    console.print("[cyan]Analyzing contract...[/]", end="")
    exports = find_exports(contract_path)
    imports = find_imports(contract_path)

    # Check for invalid export names (C keywords)
    invalid_exports = validate_export_names(exports)
    if invalid_exports:
        console.print(" [red]Error: Invalid export function names[/]")
        for invalid_export in invalid_exports:
            console.print(
                f"  [red]'{invalid_export}' is a C reserved keyword and cannot be used as an export function name[/]"
            )
        console.print("[yellow]Please rename these functions in your contract.[/]")
        sys.exit(1)

    # Show analysis results
    if not exports:
        console.print(" [yellow]No exported functions found[/]")
    else:
        console.print(
            f" found [cyan]{len(exports)}[/] exported functions: [cyan]{', '.join(sorted(exports))}[/]"
        )

    return exports, imports



try:
    from Crypto.Cipher import AES

    aes = AES.new
except ImportError:
    try:
        from cryptolib import aes
    except ImportError:
        print("SKIP")
        raise SystemExit

crypto = aes(b"1234" * 8, 2, b"5678" * 4)
enc = crypto.encrypt(bytes(range(32)))
print(enc)
crypto = aes(b"1234" * 8, 2, b"5678" * 4)
print(crypto.decrypt(enc))



"""
UART IRQ test for the CC3200 based boards.
"""

from machine import UART
import os
import time

mch = os.uname().machine
if "LaunchPad" in mch:
    uart_pins = [
        [("GP12", "GP13"), ("GP12", "GP13", "GP7", "GP6")],
        [("GP16", "GP17"), ("GP16", "GP17", "GP7", "GP6")],
    ]
elif "WiPy" in mch:
    uart_pins = [
        [("GP12", "GP13"), ("GP12", "GP13", "GP7", "GP6")],
        [("GP16", "GP17"), ("GP16", "GP17", "GP7", "GP6")],
    ]
else:
    raise Exception("Board not supported!")

# just in case we have stdio duplicated on any of the uarts
os.dupterm(None)

uart0 = UART(0, 1000000, pins=uart_pins[0][0])
uart1 = UART(1, 1000000, pins=uart_pins[1][0])

uart0_int_count = 0
uart1_int_count = 0


def uart0_handler(uart_o):
    global uart0_irq
    global uart0_int_count
    if uart0_irq.flags() & UART.RX_ANY:
        uart0_int_count += 1


def uart1_handler(uart_o):
    global uart1_irq
    global uart1_int_count
    if uart1_irq.flags() & UART.RX_ANY:
        uart1_int_count += 1


uart0_irq = uart0.irq(trigger=UART.RX_ANY, handler=uart0_handler)
uart1_irq = uart1.irq(trigger=UART.RX_ANY, handler=uart1_handler)

uart0.write(b"123")
# wait for the characters to be received
while not uart1.any():
    pass

time.sleep_us(100)
print(uart1.any() == 3)
print(uart1_int_count > 0)
print(uart1_irq.flags() == 0)
print(uart0_irq.flags() == 0)
print(uart1.read() == b"123")

uart1.write(b"12345")
# wait for the characters to be received
while not uart0.any():
    pass

time.sleep_us(100)
print(uart0.any() == 5)
print(uart0_int_count > 0)
print(uart0_irq.flags() == 0)
print(uart1_irq.flags() == 0)
print(uart0.read() == b"12345")

# do it again
uart1_int_count = 0
uart0.write(b"123")
# wait for the characters to be received
while not uart1.any():
    pass

time.sleep_us(100)
print(uart1.any() == 3)
print(uart1_int_count > 0)
print(uart1_irq.flags() == 0)
print(uart0_irq.flags() == 0)
print(uart1.read() == b"123")

# disable the interrupt
uart1_irq.disable()
# do it again
uart1_int_count = 0
uart0.write(b"123")
# wait for the characters to be received
while not uart1.any():
    pass

time.sleep_us(100)
print(uart1.any() == 3)
print(uart1_int_count == 0)  # no interrupt triggered this time
print(uart1_irq.flags() == 0)
print(uart0_irq.flags() == 0)
print(uart1.read() == b"123")

# enable the interrupt
uart1_irq.enable()
# do it again
uart1_int_count = 0
uart0.write(b"123")
# wait for the characters to be received
while not uart1.any():
    pass

time.sleep_us(100)
print(uart1.any() == 3)
print(uart1_int_count > 0)
print(uart1_irq.flags() == 0)
print(uart0_irq.flags() == 0)
print(uart1.read() == b"123")

uart1_irq.init(trigger=UART.RX_ANY, handler=None)  # No handler
# do it again
uart1_int_count = 0
uart0.write(b"123")
# wait for the characters to be received
while not uart1.any():
    pass

time.sleep_us(100)
print(uart1.any() == 3)
print(uart1_int_count == 0)  # no interrupt handler called
print(uart1_irq.flags() == 0)
print(uart0_irq.flags() == 0)
print(uart1.read() == b"123")

# check for memory leaks
for i in range(0, 1000):
    uart0_irq = uart0.irq(trigger=UART.RX_ANY, handler=uart0_handler)
    uart1_irq = uart1.irq(trigger=UART.RX_ANY, handler=uart1_handler)

# next ones must raise
try:
    uart0_irq = uart0.irq(trigger=100, handler=uart0_handler)
except:
    print("Exception")

try:
    uart0_irq = uart0.irq(trigger=0)
except:
    print("Exception")

try:
    uart0_irq = uart0.irq(trigger=UART.RX_ANY, wake=Sleep.SUSPENDED)
except:
    print("Exception")

uart0_irq.disable()
uart1_irq.disable()



# boot.py -- runs on boot-up
# Let's you choose which script to run.
# > To run 'datalogger.py':
#       * press reset and do nothing else
# > To run 'cardreader.py':
#       * press reset
#       * press user switch and hold until orange LED goes out

import pyb

pyb.LED(3).on()  # indicate we are waiting for switch press
pyb.delay(2000)  # wait for user to maybe press the switch
switch_value = pyb.Switch()()  # sample the switch at end of delay
pyb.LED(3).off()  # indicate that we finished waiting for the switch

pyb.LED(4).on()  # indicate that we are selecting the mode

if switch_value:
    pyb.usb_mode("VCP+MSC")
    pyb.main("cardreader.py")  # if switch was pressed, run this
else:
    pyb.usb_mode("VCP+HID")
    pyb.main("datalogger.py")  # if switch wasn't pressed, run this

pyb.LED(4).off()  # indicate that we finished selecting the mode



# test function args, keyword only with default value

# a single arg with a default
def f1(*, a=1):
    print(a)
f1()
f1(a=2)

# 1 arg default, 1 not
def f2(*, a=1, b):
    print(a, b)
f2(b=2)
f2(a=2, b=3)

# 1 positional, 1 arg default, 1 not
def f3(a, *, b=2, c):
    print(a, b, c)
f3(1, c=3)
f3(1, b=3, c=4)
f3(1, **{'c':3})
f3(1, **{'b':'3', 'c':4})

# many args, not all with defaults
def f4(*, a=1, b, c=3, d, e=5, f):
    print(a, b, c, d, e, f)
f4(b=2, d=4, f=6)
f4(a=11, b=2, d=4, f=6)
f4(a=11, b=2, c=33, d=4, e=55, f=6)
f4(f=6, e=55, d=4, c=33, b=2, a=11)

# positional with default, then keyword only
def f5(a, b=4, *c, d=8):
    print(a, b, c, d)
f5(1)
f5(1, d=9)
f5(1, b=44, d=9)



import os
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory


def _isgenerator(x):
    return isinstance(x, type((lambda: (yield))()))


class TestPathlib(unittest.TestCase):
    def assertExists(self, fn):
        os.stat(fn)

    def assertNotExists(self, fn):
        with self.assertRaises(OSError):
            os.stat(fn)

    def setUp(self):
        self._tmp_path_obj = TemporaryDirectory()
        self.tmp_path = self._tmp_path_obj.name

    def tearDown(self):
        self._tmp_path_obj.cleanup()

    def test_init_single_segment(self):
        path = Path("foo")
        self.assertTrue(path._path == "foo")

        path = Path("foo/")
        self.assertTrue(path._path == "foo")

        path = Path("/foo")
        self.assertTrue(path._path == "/foo")

        path = Path("/////foo")
        self.assertTrue(path._path == "/foo")

        path = Path("")
        self.assertTrue(path._path == ".")

    def test_init_multiple_segment(self):
        path = Path("foo", "bar")
        self.assertTrue(path._path == "foo/bar")

        path = Path("foo/", "bar")
        self.assertTrue(path._path == "foo/bar")

        path = Path("/foo", "bar")
        self.assertTrue(path._path == "/foo/bar")

        path = Path("/foo", "", "bar")
        self.assertTrue(path._path == "/foo/bar")

        path = Path("/foo/", "", "/bar/")
        self.assertTrue(path._path == "/bar")

        path = Path("", "")
        self.assertTrue(path._path == ".")

    def test_truediv_join_str(self):
        actual = Path("foo") / "bar"
        self.assertTrue(actual == Path("foo/bar"))

    def test_truediv_join_path(self):
        actual = Path("foo") / Path("bar")
        self.assertTrue(actual == Path("foo/bar"))

        actual = Path("foo") / Path("/bar")
        self.assertTrue(actual == "/bar")

    def test_eq_and_absolute(self):
        self.assertTrue(Path("") == Path("."))
        self.assertTrue(Path("foo") == Path(os.getcwd(), "foo"))
        self.assertTrue(Path("foo") == "foo")
        self.assertTrue(Path("foo") == os.getcwd() + "/foo")

        self.assertTrue(Path("foo") != Path("bar"))
        self.assertTrue(Path(".") != Path("/"))

    def test_open(self):
        fn = self.tmp_path + "/foo.txt"
        path = Path(fn)

        with open(fn, "w") as f:
            f.write("file contents")

        with path.open("r") as f:
            actual = f.read()

        self.assertTrue(actual == "file contents")

    def test_exists(self):
        fn = self.tmp_path + "/foo.txt"

        path = Path(str(fn))
        self.assertTrue(not path.exists())

        with open(fn, "w"):
            pass

        self.assertTrue(path.exists())

    def test_mkdir(self):
        target = self.tmp_path + "/foo/bar/baz"
        path = Path(target)

        with self.assertRaises(OSError):
            path.mkdir()

        with self.assertRaises(OSError):
            path.mkdir(exist_ok=True)

        path.mkdir(parents=True)
        self.assertExists(target)

        with self.assertRaises(OSError):
            path.mkdir(exist_ok=False)

        path.mkdir(exist_ok=True)

    def test_is_dir(self):
        target = self.tmp_path
        path = Path(target)
        self.assertTrue(path.is_dir())

        target = self.tmp_path + "/foo"
        path = Path(target)
        self.assertTrue(not path.is_dir())
        os.mkdir(target)
        self.assertTrue(path.is_dir())

        target = self.tmp_path + "/bar.txt"
        path = Path(target)
        self.assertTrue(not path.is_dir())
        with open(target, "w"):
            pass
        self.assertTrue(not path.is_dir())

    def test_is_file(self):
        target = self.tmp_path
        path = Path(target)
        self.assertTrue(not path.is_file())

        target = self.tmp_path + "/bar.txt"
        path = Path(target)
        self.assertTrue(not path.is_file())
        with open(target, "w"):
            pass
        self.assertTrue(path.is_file())

    def test_glob(self):
        foo_txt = self.tmp_path + "/foo.txt"
        with open(foo_txt, "w"):
            pass
        bar_txt = self.tmp_path + "/bar.txt"
        with open(bar_txt, "w"):
            pass
        baz_bin = self.tmp_path + "/baz.bin"
        with open(baz_bin, "w"):
            pass

        path = Path(self.tmp_path)
        glob_gen = path.glob("*.txt")
        self.assertTrue(_isgenerator(glob_gen))

        res = [str(x) for x in glob_gen]
        self.assertTrue(len(res) == 2)
        self.assertTrue(foo_txt in res)
        self.assertTrue(bar_txt in res)

    def test_rglob(self):
        foo_txt = self.tmp_path + "/foo.txt"
        with open(foo_txt, "w"):
            pass
        bar_txt = self.tmp_path + "/bar.txt"
        with open(bar_txt, "w"):
            pass
        baz_bin = self.tmp_path + "/baz.bin"
        with open(baz_bin, "w"):
            pass

        boop_folder = self.tmp_path + "/boop"
        os.mkdir(boop_folder)
        bap_txt = self.tmp_path + "/boop/bap.txt"
        with open(bap_txt, "w"):
            pass

        path = Path(self.tmp_path)
        glob_gen = path.rglob("*.txt")
        self.assertTrue(_isgenerator(glob_gen))

        res = [str(x) for x in glob_gen]
        self.assertTrue(len(res) == 3)
        self.assertTrue(foo_txt in res)
        self.assertTrue(bar_txt in res)
        self.assertTrue(bap_txt in res)

    def test_stat(self):
        expected = os.stat(self.tmp_path)
        path = Path(self.tmp_path)
        actual = path.stat()
        self.assertTrue(expected == actual)

    def test_rmdir(self):
        target = self.tmp_path + "/foo"
        path = Path(target)

        with self.assertRaises(OSError):
            # Doesn't exist
            path.rmdir()

        os.mkdir(target)
        self.assertExists(target)
        path.rmdir()
        self.assertNotExists(target)

        os.mkdir(target)
        with open(target + "/bar.txt", "w"):
            pass

        with self.assertRaises(OSError):
            # Cannot rmdir; contains file.
            path.rmdir()

    def test_touch(self):
        target = self.tmp_path + "/foo.txt"

        path = Path(target)
        path.touch()
        self.assertExists(target)

        path.touch()  # touching existing file is fine
        self.assertExists(target)

        # Technically should be FileExistsError,
        # but thats not builtin to micropython
        with self.assertRaises(OSError):
            path.touch(exist_ok=False)

        path = Path(self.tmp_path + "/bar/baz.txt")
        with self.assertRaises(OSError):
            # Parent directory does not exist
            path.touch()

    def test_unlink(self):
        target = self.tmp_path + "/foo.txt"

        path = Path(target)
        with self.assertRaises(OSError):
            # File does not exist
            path.unlink()

        with open(target, "w"):
            pass

        self.assertExists(target)
        path.unlink()
        self.assertNotExists(target)

        path = Path(self.tmp_path)
        with self.assertRaises(OSError):
            # File does not exist
            path.unlink()

    def test_write_bytes(self):
        target = self.tmp_path + "/foo.bin"
        path = Path(target)
        path.write_bytes(b"test byte data")
        with open(target, "rb") as f:
            actual = f.read()
        self.assertTrue(actual == b"test byte data")

    def test_write_text(self):
        target = self.tmp_path + "/foo.txt"
        path = Path(target)
        path.write_text("test string")
        with open(target, "r") as f:
            actual = f.read()
        self.assertTrue(actual == "test string")

    def test_read_bytes(self):
        target = self.tmp_path + "/foo.bin"
        with open(target, "wb") as f:
            f.write(b"test byte data")

        path = Path(target)
        actual = path.read_bytes()
        self.assertTrue(actual == b"test byte data")

    def test_read_text(self):
        target = self.tmp_path + "/foo.bin"
        with open(target, "w") as f:
            f.write("test string")

        path = Path(target)
        actual = path.read_text()
        self.assertTrue(actual == "test string")

    def test_stem(self):
        self.assertTrue(Path("foo/test").stem == "test")
        self.assertTrue(Path("foo/bar.bin").stem == "bar")
        self.assertTrue(Path("").stem == "")

    def test_name(self):
        self.assertTrue(Path("foo/test").name == "test")
        self.assertTrue(Path("foo/bar.bin").name == "bar.bin")

    def test_parent(self):
        self.assertTrue(Path("foo/test").parent == Path("foo"))
        self.assertTrue(Path("foo/bar.bin").parent == Path("foo"))
        self.assertTrue(Path("bar.bin").parent == Path("."))
        self.assertTrue(Path(".").parent == Path("."))
        self.assertTrue(Path("/").parent == Path("/"))

    def test_suffix(self):
        self.assertTrue(Path("foo/test").suffix == "")
        self.assertTrue(Path("foo/bar.bin").suffix == ".bin")
        self.assertTrue(Path("bar.txt").suffix == ".txt")

    def test_with_suffix(self):
        self.assertTrue(Path("foo/test").with_suffix(".tar") == Path("foo/test.tar"))
        self.assertTrue(Path("foo/bar.bin").with_suffix(".txt") == Path("foo/bar.txt"))
        self.assertTrue(Path("bar.txt").with_suffix("") == Path("bar"))

    def test_rtruediv(self):
        """Works as of micropython ea7031f"""
        res = "foo" / Path("bar")
        self.assertTrue(res == Path("foo/bar"))

    def test_rtruediv_inplace(self):
        """Works as of micropython ea7031f"""
        res = "foo"
        res /= Path("bar")
        self.assertTrue(res == Path("foo/bar"))



# test that modtls produces a numerical error message when out of heap

import socket, ssl, sys

try:
    from micropython import alloc_emergency_exception_buf, heap_lock, heap_unlock
except:
    print("SKIP")
    raise SystemExit


# test with heap locked to see it switch to number-only error message
def test(addr):
    alloc_emergency_exception_buf(256)
    s = socket.socket()
    s.connect(addr)
    try:
        s.setblocking(False)
        s = ssl.wrap_socket(s, do_handshake=False)
        heap_lock()
        print("heap is locked")
        while True:
            ret = s.write("foo")
            if ret:
                break
        heap_unlock()
        print("wrap: no exception")
    except OSError as e:
        heap_unlock()
        # mbedtls produces "-29184"
        # axtls produces "RECORD_OVERFLOW"
        ok = "-29184" in str(e) or "RECORD_OVERFLOW" in str(e)
        print("wrap:", ok)
        if not ok:
            print("got exception:", e)
    s.close()


if __name__ == "__main__":
    # connect to plain HTTP port, oops!
    addr = socket.getaddrinfo("micropython.org", 80)[0][-1]
    test(addr)



"""Additions to the TarFile class to support creating and appending tar files.

The methods defined below in are injected into the TarFile class in the
tarfile package.
"""

import uctypes
import os

# Extended subset of tar header fields including the ones we'll write.
# http://www.gnu.org/software/tar/manual/html_node/Standard.html
_TAR_HEADER = {
    "name": (uctypes.ARRAY | 0, uctypes.UINT8 | 100),
    "mode": (uctypes.ARRAY | 100, uctypes.UINT8 | 8),
    "uid": (uctypes.ARRAY | 108, uctypes.UINT8 | 8),
    "gid": (uctypes.ARRAY | 116, uctypes.UINT8 | 8),
    "size": (uctypes.ARRAY | 124, uctypes.UINT8 | 12),
    "mtime": (uctypes.ARRAY | 136, uctypes.UINT8 | 12),
    "chksum": (uctypes.ARRAY | 148, uctypes.UINT8 | 8),
    "typeflag": (uctypes.ARRAY | 156, uctypes.UINT8 | 1),
}


_NUL = const(b"\0")  # the null character
_BLOCKSIZE = const(512)  # length of processing blocks
_RECORDSIZE = const(_BLOCKSIZE * 20)  # length of records


def _open_write(self, name, mode, fileobj):
    if mode == "w":
        if not fileobj:
            self.f = open(name, "wb")
        else:
            self.f = fileobj
    elif mode == "a":
        if not fileobj:
            self.f = open(name, "r+b")
        else:
            self.f = fileobj
        # Read through the existing file.
        while self.next():
            pass
        # Position at start of end block.
        self.f.seek(self.offset)
    else:
        raise ValueError("mode " + mode + " not supported.")


def _close_write(self):
    # Must be called to complete writing a tar file.
    if self.mode == "w":
        self.f.write(_NUL * (_BLOCKSIZE * 2))
        self.offset += _BLOCKSIZE * 2
        remainder = self.offset % _RECORDSIZE
        if remainder:
            self.f.write(_NUL * (_RECORDSIZE - remainder))


def addfile(self, tarinfo, fileobj=None):
    # Write the header: 100 bytes of name, 8 bytes of mode in octal...
    buf = bytearray(_BLOCKSIZE)
    name = tarinfo.name
    size = tarinfo.size
    if tarinfo.isdir():
        size = 0
        if not name.endswith("/"):
            name += "/"
    hdr = uctypes.struct(uctypes.addressof(buf), _TAR_HEADER, uctypes.LITTLE_ENDIAN)
    hdr.name[:] = name.encode("utf-8")[:100]
    hdr.mode[:] = b"%07o\0" % ((0o755 if tarinfo.isdir() else 0o644) & 0o7777)
    hdr.uid[:] = b"%07o\0" % tarinfo.uid
    hdr.gid[:] = b"%07o\0" % tarinfo.gid
    hdr.size[:] = b"%011o\0" % size
    hdr.mtime[:] = b"%011o\0" % tarinfo.mtime
    hdr.typeflag[:] = b"5" if tarinfo.isdir() else b"0"
    # Checksum is calculated with checksum field all blanks.
    hdr.chksum[:] = b"        "
    # Calculate and insert the actual checksum.
    chksum = sum(buf)
    hdr.chksum[:] = b"%06o\0 " % chksum
    # Emit the header.
    self.f.write(buf)
    self.offset += len(buf)

    # Copy the file contents, if any.
    if fileobj:
        n_bytes = self.f.write(fileobj.read())
        self.offset += n_bytes
        remains = -n_bytes & (_BLOCKSIZE - 1)  # == 0b111111111
        if remains:
            buf = bytearray(remains)
            self.f.write(buf)
            self.offset += len(buf)


def add(self, name, recursive=True):
    from . import TarInfo

    try:
        stat = os.stat(name)
        res_name = (name + '/') if (stat[0] & 0xf000) == 0x4000 else name
        tarinfo = TarInfo(res_name)
        tarinfo.mode = stat[0]
        tarinfo.uid = stat[4]
        tarinfo.gid = stat[5]
        tarinfo.size = stat[6]
        tarinfo.mtime = stat[8]
    except OSError:
        print("Cannot stat", name, " - skipping.")
        return
    if not (tarinfo.isdir() or tarinfo.isreg()):
        # We only accept directories or regular files.
        print(name, "is not a directory or regular file - skipping.")
        return
    if tarinfo.isdir():
        self.addfile(tarinfo)
        if recursive:
            for f in os.ilistdir(name):
                self.add(name + "/" + f[0], recursive)
    else:  # type == REGTYPE
        self.addfile(tarinfo, open(name, "rb"))



import time
from signal import *

quit = 0


def handler(signum):
    global quit
    quit = 1
    print("Signal handler called with signal", signum)


print("org signal() val:", signal(SIGINT, handler))
print("read back signal() val:", signal(SIGINT, handler))

while not quit:
    time.sleep(10)

print("Quitting...")



# MicroPython lora reliable_delivery example - asynchronous sender program
# MIT license; Copyright (c) 2023 Angus Gratton
import machine
from machine import SPI, Pin
import random
import struct
import time
import asyncio

from lora_rd_settings import RECEIVER_ID, ACK_LENGTH, ACK_DELAY_MS, lora_cfg

SLEEP_BETWEEN_MS = 5000  # Main loop should sleep this long between sending data to the receiver

MAX_RETRIES = 4  # Retry each message this often if no ACK is received

# Initial retry is after this long. Increases by 1.25x each subsequent retry.
BASE_RETRY_TIMEOUT_MS = 1000

# Add random jitter to each retry period, up to this long. Useful to prevent two
# devices ending up in sync.
RETRY_JITTER_MS = 1500

# If reported RSSI value is lower than this, increase
# output power 1dBm
RSSI_WEAK_THRESH = -110

# If reported RSSI value is higher than this, decrease
# output power 1dBm
RSSI_STRONG_THRESH = -70

# IMPORTANT: Set this to the maximum output power in dBm that is permitted in
# your regulatory environment.
OUTPUT_MAX_DBM = 15
OUTPUT_MIN_DBM = -20


def get_async_modem():
    # from lora import AsyncSX1276
    # return AsyncSX1276(
    #     spi=SPI(1, baudrate=2000_000, polarity=0, phase=0,
    #             miso=Pin(19), mosi=Pin(27), sck=Pin(5)),
    #     cs=Pin(18),
    #     dio0=Pin(26),
    #     dio1=Pin(35),
    #     reset=Pin(14),
    #     lora_cfg=lora_cfg,
    # )
    raise NotImplementedError("Replace this function with one that returns a lora modem instance")


def main():
    modem = get_async_modem()
    asyncio.run(sender_task(modem))


async def sender_task(modem):
    # Unique ID of this sender, 16-bit number. This method of generating an ID is pretty crummy,
    # if using this in a real application then probably better to store these in the filesystem or
    # something like that
    DEVICE_ID = sum(b for b in machine.unique_id()) & 0xFFFF

    sender = AsyncSender(modem, DEVICE_ID)
    while True:
        sensor_data = await get_sensor_data()
        await sender.send(sensor_data)

        # Sleep until the next time we should read the sensor data and send it to
        # the receiver. awaiting here means other tasks will run.
        modem.sleep()
        await asyncio.sleep_ms(SLEEP_BETWEEN_MS)


async def get_sensor_data():
    # Return a bytes object with the latest sensor data to send to the receiver.
    #
    # As this is just an example, we send a dummy payload which is just a string
    # containing our ticks_ms() timestamp.
    #
    # In a real application the sensor data should usually be binary data and
    # not a string, to save transmission size.
    return f"Hello, ticks_ms={time.ticks_ms()}".encode()


class AsyncSender:
    def __init__(self, modem, device_id):
        self.modem = modem
        self.device_id = device_id
        self.counter = 0
        self.output_power = lora_cfg["output_power"]  # start with common settings power level
        self.rx_ack = None  # reuse the ack message object when we can

        print(f"Sender initialized with ID {device_id:#x}")
        random.seed(device_id)
        self.adjust_output_power(0)  # set the initial value within MIN/MAX

        modem.calibrate()

    async def send(self, sensor_data, adjust_output_power=True):
        # Send a packet of sensor data to the receiver reliably.
        #
        # Returns True if data was successfully sent and ACKed, False otherwise.
        #
        # If adjust_output_power==True then increase or decrease output power
        # according to the RSSI reported in the ACK packet.
        self.counter = (self.counter + 1) & 0xFF

        # Prepare the simple payload with header and checksum
        # See README for a summary of the simple data message format
        payload = bytearray(len(sensor_data) + 5)
        struct.pack_into("<HBB", payload, 0, self.device_id, self.counter, len(sensor_data))
        payload[4:-1] = sensor_data
        payload[-1] = sum(b for b in payload) & 0xFF

        # Calculate the time on air (in milliseconds) for an ACK packet
        ack_packet_ms = self.modem.get_time_on_air_us(ACK_LENGTH) // 1000 + 1
        timeout = BASE_RETRY_TIMEOUT_MS

        print(f"Sending {len(payload)} bytes")

        # Send the payload, until we receive an acknowledgement or run out of retries
        for _ in range(MAX_RETRIES):
            sent_at = await self.modem.send(payload)

            # We expect the receiver of a valid message to start sending the ACK
            # approximately ACK_DELAY_MS after receiving the message (to allow
            # the sender time to reconfigure the modem.)
            #
            # We start receiving as soon as we can, but allow up to
            # ACK_DELAY_MS*2 of total timing leeway - plus the time on air for
            # the packet itself
            maybe_ack = await self.modem.recv(
                ack_packet_ms + ACK_DELAY_MS * 2, rx_packet=self.rx_ack
            )

            # Check if the packet we received is a valid ACK
            rssi = self._ack_is_valid(maybe_ack, payload[-1])

            if rssi is not None:  # ACK is valid
                self.rx_ack == maybe_ack

                delta = time.ticks_diff(maybe_ack.ticks_ms, sent_at)
                print(
                    f"ACKed with RSSI {rssi}, {delta}ms after sent "
                    + f"(skew {delta-ACK_DELAY_MS-ack_packet_ms}ms)"
                )

                if adjust_output_power:
                    if rssi > RSSI_STRONG_THRESH:
                        self.adjust_output_power(-1)
                    elif rssi < RSSI_WEAK_THRESH:
                        self.adjust_output_power(1)

                return True

            # Otherwise, prepare to sleep briefly and then retry
            next_try_at = time.ticks_add(sent_at, timeout)
            sleep_time = time.ticks_diff(next_try_at, time.ticks_ms()) + random.randrange(
                RETRY_JITTER_MS
            )
            if sleep_time > 0:
                self.modem.sleep()
                await asyncio.sleep_ms(sleep_time)

            # add 25% timeout for next iteration
            timeout = (timeout * 5) // 4

        print(f"Failed, no ACK after {MAX_RETRIES} retries.")
        if adjust_output_power:
            self.adjust_output_power(2)
        self.modem.calibrate_image()  # try and improve the RX sensitivity for next time
        return False

    def _ack_is_valid(self, maybe_ack, csum):
        # Private function to verify if the RxPacket held in 'maybe_ack' is a valid ACK for the
        # current device_id and counter value, and provided csum value.
        #
        # If it is, returns the reported RSSI value from the packet.
        # If not, returns None
        if (not maybe_ack) or len(maybe_ack) != ACK_LENGTH:
            return None

        base_id, ack_id, ack_counter, ack_csum, rssi = struct.unpack("<HHBBb", maybe_ack)

        if (
            base_id != RECEIVER_ID
            or ack_id != self.device_id
            or ack_counter != self.counter
            or ack_csum != csum
        ):
            return None

        return rssi

    def adjust_output_power(self, delta_dbm):
        # Adjust the modem output power by +/-delta_dbm, max of OUTPUT_MAX_DBM
        #
        # (note: the radio may also apply its own power limit internally.)
        new = max(min(self.output_power + delta_dbm, OUTPUT_MAX_DBM), OUTPUT_MIN_DBM)
        self.output_power = new
        print(f"New output_power {new}/{OUTPUT_MAX_DBM} (delta {delta_dbm})")
        self.modem.configure({"output_power": self.output_power})


if __name__ == "__main__":
    main()



# Tests domain errors in math functions.
# This is split out from math_domain.py because math.pow(0, -inf) was changed
# in Python 3.11, and so this test requires a .py.exp file.
# (See https://github.com/python/cpython/issues/88505)

try:
    import math
except ImportError:
    print("SKIP")
    raise SystemExit

inf = float("inf")

for name, f in (
    ("pow", math.pow),
    ("log", math.log),
    ("fmod", math.fmod),
    ("atan2", math.atan2),
    ("copysign", math.copysign),
):
    for x in ((0, -inf),):
        try:
            ans = "%.4f" % f(*x)
        except ValueError:
            ans = "ValueError"
        print("%s(%.4f, %.4f) = %s" % (name, x[0], x[1], ans))



try:
    import framebuf, sys
except ImportError:
    print("SKIP")
    raise SystemExit

# This test and its .exp file is based on a little-endian architecture.
if sys.byteorder != "little":
    print("SKIP")
    raise SystemExit


def printbuf():
    print("--8<--")
    for y in range(h):
        print(buf[y * w * 2 : (y + 1) * w * 2])
    print("-->8--")


w = 4
h = 5
buf = bytearray(w * h * 2)
fbuf = framebuf.FrameBuffer(buf, w, h, framebuf.RGB565)

# fill
fbuf.fill(0xFFFF)
printbuf()
fbuf.fill(0x0000)
printbuf()

# put pixel
fbuf.pixel(0, 0, 0xEEEE)
fbuf.pixel(3, 0, 0xEE00)
fbuf.pixel(0, 4, 0x00EE)
fbuf.pixel(3, 4, 0x0EE0)
printbuf()

# get pixel
print(fbuf.pixel(0, 4), fbuf.pixel(1, 1))

# scroll
fbuf.fill(0x0000)
fbuf.pixel(2, 2, 0xFFFF)
printbuf()
fbuf.scroll(0, 1)
printbuf()
fbuf.scroll(1, 0)
printbuf()
fbuf.scroll(-1, -2)
printbuf()

w2 = 2
h2 = 3
buf2 = bytearray(w2 * h2 * 2)
fbuf2 = framebuf.FrameBuffer(buf2, w2, h2, framebuf.RGB565)

fbuf2.fill(0x0000)
fbuf2.pixel(0, 0, 0x0EE0)
fbuf2.pixel(0, 2, 0xEE00)
fbuf2.pixel(1, 0, 0x00EE)
fbuf2.pixel(1, 2, 0xE00E)
fbuf.fill(0xFFFF)
fbuf.blit(fbuf2, 3, 3, 0x0000)
fbuf.blit(fbuf2, -1, -1, 0x0000)
fbuf.blit(fbuf2, 16, 16, 0x0000)
printbuf()



from pyb import ADC, Timer

adct = ADC(16)  # Temperature 930 -> 20C
print(str(adct)[:19])
adcv = ADC(17)  # Voltage 1500 -> 3.3V
print(adcv)

# read single sample; 2.5V-5V is pass range
val = adcv.read()
assert val > 1000 and val < 2000

# timer for read_timed
tim = Timer(5, freq=500)

# read into bytearray
buf = bytearray(b"\xff" * 50)
adcv.read_timed(buf, tim)
print(len(buf))
for i in buf:
    assert i > 50 and i < 150

# read into arrays with different element sizes
import array

arv = array.array("h", 25 * [0x7FFF])
adcv.read_timed(arv, tim)
print(len(arv))
for i in arv:
    assert i > 1000 and i < 2000

arv = array.array("i", 30 * [-1])
adcv.read_timed(arv, tim)
print(len(arv))
for i in arv:
    assert i > 1000 and i < 2000

# Test read_timed_multi
arv = bytearray(b"\xff" * 50)
art = bytearray(b"\xff" * 50)
ADC.read_timed_multi((adcv, adct), (arv, art), tim)
for i in arv:
    assert i > 60 and i < 125
# Wide range: unsure of accuracy of temp sensor.
for i in art:
    assert i > 15 and i < 200

arv = array.array("i", 25 * [-1])
art = array.array("i", 25 * [-1])
ADC.read_timed_multi((adcv, adct), (arv, art), tim)
for i in arv:
    assert i > 1000 and i < 2000
# Wide range: unsure of accuracy of temp sensor.
for i in art:
    assert i > 50 and i < 2000

arv = array.array("h", 25 * [0x7FFF])
art = array.array("h", 25 * [0x7FFF])
ADC.read_timed_multi((adcv, adct), (arv, art), tim)
for i in arv:
    assert i > 1000 and i < 2000
# Wide range: unsure of accuracy of temp sensor.
for i in art:
    assert i > 50 and i < 2000



#
# Driver class for the WM8960 Codec to be used e.g. with MIMXRT_1xxx Boards.
# Derived from the NXP SDK drivers.
#
# Copyright (c) 2015, Freescale Semiconductor, Inc., (C-Code)
# Copyright 2016-2021 NXP, (C-Code)
# All rights reserved.
#
# Translated to MicroPython by Robert Hammelrath, 2022
#
# SPDX-License-Identifier: BSD-3-Clause
#

import array
from micropython import const

# Define the register addresses of WM8960.
_LINVOL = const(0x0)
_RINVOL = const(0x1)
_LOUT1 = const(0x2)
_ROUT1 = const(0x3)
_CLOCK1 = const(0x4)
_DACCTL1 = const(0x5)
_DACCTL2 = const(0x6)
_IFACE1 = const(0x7)
_CLOCK2 = const(0x8)
_IFACE2 = const(0x9)
_LDAC = const(0xA)
_RDAC = const(0xB)
_RESET = const(0xF)
_3D = const(0x10)
_ALC1 = const(0x11)
_ALC2 = const(0x12)
_ALC3 = const(0x13)
_NOISEG = const(0x14)
_LADC = const(0x15)
_RADC = const(0x16)
_ADDCTL1 = const(0x17)
# Register _ADDCTL2 = const(0x18)
_POWER1 = const(0x19)
_POWER2 = const(0x1A)
_ADDCTL3 = const(0x1B)
# Register _APOP1 = const(0x1C)
# Register _APOP2 = const(0x1D)
_LINPATH = const(0x20)
_RINPATH = const(0x21)
_LOUTMIX = const(0x22)
_ROUTMIX = const(0x25)
_MONOMIX1 = const(0x26)
_MONOMIX2 = const(0x27)
_LOUT2 = const(0x28)
_ROUT2 = const(0x29)
_MONO = const(0x2A)
_INBMIX1 = const(0x2B)
_INBMIX2 = const(0x2C)
_BYPASS1 = const(0x2D)
_BYPASS2 = const(0x2E)
_POWER3 = const(0x2F)
_ADDCTL4 = const(0x30)
_CLASSD1 = const(0x31)
# Register _CLASSD3 = const(0x33)
_PLL1 = const(0x34)
_PLL2 = const(0x35)
_PLL3 = const(0x36)
_PLL4 = const(0x37)

# WM8960 PLLN range */
_PLL_N_MIN_VALUE = const(6)
_PLL_N_MAX_VALUE = const(12)

# WM8960 CLOCK2 bits
_CLOCK2_BCLK_DIV_MASK = const(0x0F)
_CLOCK2_DCLK_DIV_MASK = const(0x1C0)
_CLOCK2_DCLK_DIV_SHIFT = const(0x06)

# Register _IFACE1
_IFACE1_FORMAT_MASK = const(0x03)
_IFACE1_WL_MASK = const(0x0C)
_IFACE1_WL_SHIFT = const(0x02)
_IFACE1_LRP_MASK = const(0x10)
_IFACE1_MS_MASK = const(0x40)
_IFACE1_DLRSWAP_MASK = const(0x20)
_IFACE1_ALRSWAP_MASK = const(0x100)

# Register _POWER1
_POWER1_VREF_MASK = const(0x40)
_POWER1_VREF_SHIFT = const(0x06)
_POWER1_AINL_MASK = const(0x20)
_POWER1_AINR_MASK = const(0x10)
_POWER1_ADCL_MASK = const(0x08)
_POWER1_ADCR_MASK = const(0x0)
_POWER1_MICB_MASK = const(0x02)
_POWER1_MICB_SHIFT = const(0x01)

# Register _POWER2
_POWER2_DACL_MASK = const(0x100)
_POWER2_DACR_MASK = const(0x80)
_POWER2_LOUT1_MASK = const(0x40)
_POWER2_ROUT1_MASK = const(0x20)
_POWER2_SPKL_MASK = const(0x10)
_POWER2_SPKR_MASK = const(0x08)
_POWER3_LMIC_MASK = const(0x20)
_POWER3_RMIC_MASK = const(0x10)
_POWER3_LOMIX_MASK = const(0x08)
_POWER3_ROMIX_MASK = const(0x04)

# Register _DACCTL1 .. 3
_DACCTL1_MONOMIX_MASK = const(0x10)
_DACCTL1_MONOMIX_SHIFT = const(0x4)
_DACCTL1_DACMU_MASK = const(0x08)
_DACCTL1_DEEM_MASK = const(0x06)
_DACCTL1_DEEM_SHIFT = const(0x01)
_DACCTL2_DACSMM_MASK = const(0x08)
_DACCTL2_DACMR_MASK = const(0x04)
_DACCTL3_ALCSR_MASK = const(0x07)

# _WM8060_ALC1 .. 3
_ALC_CHANNEL_MASK = const(0x180)
_ALC_CHANNEL_SHIFT = const(0x7)
_ALC_MODE_MASK = const(0x100)
_ALC_MODE_SHIFT = const(0x8)
_ALC_GAIN_MASK = const(0x70)
_ALC_GAIN_SHIFT = const(0x4)
_ALC_TARGET_MASK = const(0x0F)
_ALC_ATTACK_MASK = const(0x0F)
_ALC_DECAY_MASK = const(0xF0)
_ALC_DECAY_SHIFT = const(4)
_ALC_HOLD_MASK = const(0xF)

# Register _NOISEG
_NOISEG_LEVEL_SHIFT = const(3)

_I2C_ADDR = const(0x1A)

# WM8960 maximum volume values
_MAX_VOLUME_ADC = const(0xFF)
_MAX_VOLUME_DAC = const(0xFF)
_MAX_VOLUME_HEADPHONE = const(0x7F)
_MAX_VOLUME_LINEIN = const(0x3F)
_MAX_VOLUME_SPEAKER = const(0x7F)

# Config symbol names
# Modules
MODULE_ADC = const(0)  # ADC module in WM8960
MODULE_DAC = const(1)  # DAC module in WM8960
MODULE_VREF = const(2)  # VREF module
MODULE_HEADPHONE = const(3)  # Headphone
MODULE_MIC_BIAS = const(4)  # Mic bias
MODULE_MIC = const(5)  # Input Mic
MODULE_LINE_IN = const(6)  # Analog in PGA
MODULE_LINE_OUT = const(7)  # Line out module
MODULE_SPEAKER = const(8)  # Speaker module
MODULE_OMIX = const(9)  # Output mixer
MODULE_MONO_OUT = const(10)  # Mono mix

# Route
ROUTE_BYPASS = const(0)  # LINEIN->Headphone.
ROUTE_PLAYBACK = const(1)  #  I2SIN->DAC->Headphone.
ROUTE_PLAYBACK_RECORD = const(2)  # I2SIN->DAC->Headphone, LINEIN->ADC->I2SOUT.
ROUTE_RECORD = const(5)  # LINEIN->ADC->I2SOUT.

# Input
INPUT_CLOSED = const(0)  # Input device is closed
INPUT_MIC1 = const(1)  # Input as single ended mic, only use L/RINPUT1
INPUT_MIC2 = const(2)  # Input as diff. mic, use L/RINPUT1 and L/RINPUT2
INPUT_MIC3 = const(3)  # Input as diff. mic, use L/RINPUT1 and L/RINPUT3
INPUT_LINE2 = const(4)  # Input as line input, only use L/RINPUT2
INPUT_LINE3 = const(5)  # Input as line input, only use L/RINPUT3

# ADC sync input
SYNC_ADC = const(0)  # Use ADCLRC pin for ADC sync
SYNC_DAC = const(1)  # used DACLRC pin for ADC sync

# Protocol type
BUS_I2S = const(2)  # I2S type
BUS_LEFT_JUSTIFIED = const(1)  # Left justified mode
BUS_RIGHT_JUSTIFIED = const(0)  # Right justified mode
BUS_PCMA = const(3)  # PCM A mode
BUS_PCMB = const(3 | (1 << 4))  # PCM B mode

# Channel swap
SWAP_NONE = const(0)
SWAP_INPUT = const(1)
SWAP_OUTPUT = const(2)

# Mute settings
MUTE_FAST = const(0)
MUTE_SLOW = const(1)

# ALC settings
ALC_OFF = const(0)
ALC_RIGHT = const(1)
ALC_LEFT = const(2)
ALC_STEREO = const(3)
ALC_MODE = const(0)  # ALC mode
ALC_LIMITER = const(1)  # Limiter mode

# Clock Source
SYSCLK_MCLK = const(0)  # sysclk source from external MCLK
SYSCLK_PLL = const(1)  # sysclk source from internal PLL


class Regs:
    # register cache of 56 register. Since registers cannot be read back, they are
    # kept in the table for modification
    # fmt: off
    cache = array.array("H", (
        0x0097, 0x0097, 0x0000, 0x0000, 0x0000, 0x0008, 0x0000,
        0x000a, 0x01c0, 0x0000, 0x00ff, 0x00ff, 0x0000, 0x0000,
        0x0000, 0x0000, 0x0000, 0x007b, 0x0100, 0x0032, 0x0000,
        0x00c3, 0x00c3, 0x01c0, 0x0000, 0x0000, 0x0000, 0x0000,
        0x0000, 0x0000, 0x0000, 0x0000, 0x0100, 0x0100, 0x0050,
        0x0050, 0x0050, 0x0050, 0x0000, 0x0000, 0x0000, 0x0000,
        0x0040, 0x0000, 0x0000, 0x0050, 0x0050, 0x0000, 0x0002,
        0x0037, 0x004d, 0x0080, 0x0008, 0x0031, 0x0026, 0x00e9
    ))
    # fmt: on

    def __init__(self, i2c, i2c_address=_I2C_ADDR):
        self.value_buffer = bytearray(2)
        self.i2c = i2c
        self.i2c_address = i2c_address

    def __getitem__(self, reg):
        return self.cache[reg]

    def __setitem__(self, reg, value):
        if type(reg) is tuple:
            if type(value) is tuple:
                self[reg[0]] = value[0]
                self[reg[1]] = value[1]
            else:
                self[reg[0]] = value
                self[reg[1]] = value
        else:
            if type(value) is tuple:
                val = (self.cache[reg] & (~value[0] & 0xFFFF)) | value[1]
            else:
                val = value
            self.cache[reg] = val
            self.value_buffer[0] = (reg << 1) | ((val >> 8) & 0x01)
            self.value_buffer[1] = val & 0xFF
            self.i2c.writeto(self.i2c_address, self.value_buffer)


class WM8960:
    _bit_clock_divider_table = {
        2: 0,
        3: 1,
        4: 2,
        6: 3,
        8: 4,
        11: 5,
        12: 6,
        16: 7,
        22: 8,
        24: 9,
        32: 10,
        44: 11,
        48: 12,
    }

    _dac_divider_table = {
        1.0 * 256: 0b000,
        1.5 * 256: 0b001,
        2 * 256: 0b010,
        3 * 256: 0b011,
        4 * 256: 0b100,
        5.5 * 256: 0b101,
        6 * 256: 0b110,
    }

    _audio_word_length_table = {
        16: 0b00,
        20: 0b01,
        24: 0b10,
        32: 0b11,
    }

    _alc_sample_rate_table = {
        48000: 0,
        44100: 0,
        32000: 1,
        24000: 2,
        22050: 2,
        16000: 3,
        12000: 4,
        11025: 4,
        8000: 5,
    }

    _volume_config_table = {
        MODULE_ADC: (_MAX_VOLUME_ADC, _LADC, 0x100),
        MODULE_DAC: (_MAX_VOLUME_DAC, _LDAC, 0x100),
        MODULE_HEADPHONE: (_MAX_VOLUME_HEADPHONE, _LOUT1, 0x180),
        MODULE_LINE_IN: (_MAX_VOLUME_LINEIN, _LINVOL, 0x140),
        MODULE_SPEAKER: (_MAX_VOLUME_SPEAKER, _LOUT2, 0x180),
    }

    _input_config_table = {
        INPUT_CLOSED: None,
        INPUT_MIC1: (0x138, 0x117),
        INPUT_MIC2: (0x178, 0x117),
        INPUT_MIC3: (0x1B8, 0x117),
        INPUT_LINE2: (0, 0xE),
        INPUT_LINE3: (0, 0x70),
    }

    def __init__(
        self,
        i2c,
        sample_rate=16000,
        bits=16,
        swap=SWAP_NONE,
        route=ROUTE_PLAYBACK_RECORD,
        left_input=INPUT_MIC3,
        right_input=INPUT_MIC2,
        sysclk_source=SYSCLK_MCLK,
        mclk_freq=None,
        primary=False,
        adc_sync=SYNC_DAC,
        protocol=BUS_I2S,
        i2c_address=_I2C_ADDR,
    ):
        self.regs = regs = Regs(i2c, i2c_address)
        self.sample_rate = sample_rate

        # check parameter consistency and set the sysclk value
        if sysclk_source == SYSCLK_PLL:
            if sample_rate in (11025, 22050, 44100):
                sysclk = 11289600
            else:
                sysclk = 12288000
            if sysclk < sample_rate * 256:
                sysclk = sample_rate * 256
            if mclk_freq is None:
                mclk_freq = sysclk
        else:  # sysclk_source == SYSCLK_MCLK
            if mclk_freq is None:
                mclk_freq = sample_rate * 256
            sysclk = mclk_freq

        regs[_RESET] = 0x00
        # VMID=50K, Enable VREF, AINL, AINR, ADCL and ADCR
        # I2S_IN (bit 0), I2S_OUT (bit 1), DAP (bit 4), DAC (bit 5), ADC (bit 6) are powered on
        regs[_POWER1] = 0xFE
        # Enable DACL, DACR, LOUT1, ROUT1, PLL down, SPKL, SPKR
        regs[_POWER2] = 0x1F8
        # Enable left and right channel input PGA, left and right output mixer
        regs[_POWER3] = 0x3C

        if adc_sync == SYNC_ADC:
            # ADC and DAC use different Frame Clock Pins
            regs[_IFACE2] = 0x00  # ADCLRC 0x00:Input 0x40:output.
        else:
            # ADC and DAC use the same Frame Clock Pin
            regs[_IFACE2] = 0x40  # ADCLRC 0x00:Input 0x40:output.
        self.set_data_route(route)
        self.set_protocol(protocol)

        if sysclk_source == SYSCLK_PLL:
            self.set_internal_pll_config(mclk_freq, sysclk)
        if primary:
            self.set_master_clock(sysclk, sample_rate, bits)
            # set master bit.
            self.regs[_IFACE1] = (0, _IFACE1_MS_MASK)

        self.set_speaker_clock(sysclk)

        # swap channels
        if swap & SWAP_INPUT:
            regs[_IFACE1] = (0, _IFACE1_ALRSWAP_MASK)
        if swap & SWAP_OUTPUT:
            regs[_IFACE1] = (0, _IFACE1_DLRSWAP_MASK)

        self.set_left_input(left_input)
        self.set_right_input(right_input)

        regs[_ADDCTL1] = 0x0C0
        regs[_ADDCTL4] = 0x60  # Set GPIO1 to 0.

        regs[_BYPASS1] = regs[_BYPASS2] = 0x0
        # ADC volume, 0dB
        regs[_LADC, _RADC] = 0x1C3
        # Digital DAC volume, 0dB
        regs[_LDAC, _RDAC] = 0x1FF
        # Headphone volume, LOUT1 and ROUT1, 0dB
        regs[_LOUT1, _ROUT1] = 0x16F
        # speaker volume 6dB
        regs[_LOUT2, _ROUT2] = 0x1FF
        # enable class D output
        regs[_CLASSD1] = 0xF7
        # Unmute DAC.
        regs[_DACCTL1] = 0x0000
        # Input PGA volume 0 dB
        regs[_LINVOL, _RINVOL] = 0x117

        self.config_data_format(sysclk, sample_rate, bits)

    def deinit(self):
        self.set_module(MODULE_ADC, False)
        self.set_module(MODULE_DAC, False)
        self.set_module(MODULE_VREF, False)
        self.set_module(MODULE_LINE_IN, False)
        self.set_module(MODULE_LINE_OUT, False)
        self.set_module(MODULE_SPEAKER, False)

    def set_internal_pll_config(self, input_mclk, output_clk):
        regs = self.regs
        pllF2 = output_clk * 4
        pll_prescale = 0
        sysclk_div = 1
        frac_mode = 0

        # disable PLL power
        regs[_POWER2] = (1, 0)
        regs[_CLOCK1] = (7, 0)

        pllN = pllF2 // input_mclk
        if pllN < _PLL_N_MIN_VALUE:
            input_mclk //= 2
            pll_prescale = 1
            pllN = pllF2 // input_mclk
            if pllN < _PLL_N_MIN_VALUE:
                sysclk_div = 2
                pllF2 *= 2
                pllN = pllF2 // input_mclk

        if (pllN < _PLL_N_MIN_VALUE) or (pllN > _PLL_N_MAX_VALUE):
            raise ValueError("Invalid MCLK vs. sysclk ratio")

        pllK = ((pllF2 % input_mclk) * (1 << 24)) // input_mclk
        if pllK != 0:
            frac_mode = 1

        regs[_PLL1] = (frac_mode << 5) | (pll_prescale << 4) | (pllN & 0x0F)
        regs[_PLL2] = (pllK >> 16) & 0xFF
        regs[_PLL3] = (pllK >> 8) & 0xFF
        regs[_PLL4] = pllK & 0xFF
        # enable PLL power
        regs[_POWER2] = (1, 1)
        regs[_CLOCK1] = (7, ((0 if sysclk_div == 1 else sysclk_div) << 1) | 1)

    def set_master_clock(self, sysclk, sample_rate, bit_width):
        bit_clock_divider = (sysclk * 2) // (sample_rate * bit_width * 2)
        try:
            reg_divider = self._bit_clock_divider_table[bit_clock_divider]
        except:
            raise ValueError("Invalid ratio of sysclk sample rate and bits")
        # configure the master bit clock divider
        self.regs[_CLOCK2] = (_CLOCK2_BCLK_DIV_MASK, reg_divider)

    def set_speaker_clock(self, sysclk):
        speaker_divider_table = (1.5, 2, 3, 4, 6, 8, 12, 16)
        for val in range(8):
            divider = speaker_divider_table[val]
            f = sysclk / divider
            if 500_000 < f < 1_000_000:
                break
        else:
            val = 7
        self.regs[_CLOCK2] = (
            _CLOCK2_DCLK_DIV_MASK,
            val << _CLOCK2_DCLK_DIV_SHIFT,
        )

    def set_module(self, module, is_enabled):
        is_enabled = 1 if is_enabled else 0
        regs = self.regs

        if module == MODULE_ADC:
            regs[_POWER1] = (
                _POWER1_ADCL_MASK | _POWER1_ADCR_MASK,
                (_POWER1_ADCL_MASK | _POWER1_ADCR_MASK) * is_enabled,
            )

        elif module == MODULE_DAC:
            regs[_POWER2] = (
                _POWER2_DACL_MASK | _POWER2_DACR_MASK,
                (_POWER2_DACL_MASK | _POWER2_DACR_MASK) * is_enabled,
            )

        elif module == MODULE_VREF:
            regs[_POWER1] = (
                _POWER1_VREF_MASK,
                (is_enabled << _POWER1_VREF_SHIFT),
            )

        elif module == MODULE_LINE_IN:
            regs[_POWER1] = (
                _POWER1_AINL_MASK | _POWER1_AINR_MASK,
                (_POWER1_AINL_MASK | _POWER1_AINR_MASK) * is_enabled,
            )
            regs[_POWER3] = (
                _POWER3_LMIC_MASK | _POWER3_RMIC_MASK,
                (_POWER3_LMIC_MASK | _POWER3_RMIC_MASK) * is_enabled,
            )

        elif module == MODULE_LINE_OUT:
            regs[_POWER2] = (
                _POWER2_LOUT1_MASK | _POWER2_ROUT1_MASK,
                (_POWER2_LOUT1_MASK | _POWER2_ROUT1_MASK) * is_enabled,
            )

        elif module == MODULE_MIC_BIAS:
            regs[_POWER1] = (
                _POWER1_MICB_MASK,
                (is_enabled << _POWER1_MICB_SHIFT),
            )

        elif module == MODULE_SPEAKER:
            regs[_POWER2] = (
                _POWER2_SPKL_MASK | _POWER2_SPKR_MASK,
                (_POWER2_SPKL_MASK | _POWER2_SPKR_MASK) * is_enabled,
            )
            regs[_CLASSD1] = 0xF7

        elif module == MODULE_OMIX:
            regs[_POWER3] = (
                _POWER3_LOMIX_MASK | _POWER3_ROMIX_MASK,
                (_POWER3_LOMIX_MASK | _POWER3_ROMIX_MASK) * is_enabled,
            )

        elif module == MODULE_MONO_OUT:
            regs[_MONOMIX1] = regs[_MONOMIX2] = is_enabled << 7
            regs[_MONO] = is_enabled << 6

        else:
            raise ValueError("Invalid module")

    def enable_module(self, module):
        self.set_module(module, True)

    def disable_module(self, module):
        self.set_module(module, False)

    def set_data_route(self, route):
        regs = self.regs
        if route == ROUTE_BYPASS:
            # Bypass means from line-in to HP
            # Left LINPUT3 to left output mixer, LINPUT3 left output mixer volume = 0dB
            # Right RINPUT3 to right output mixer, RINPUT3 right output mixer volume = 0dB
            regs[_LOUTMIX, _ROUTMIX] = 0x80

        elif route == ROUTE_PLAYBACK:
            # Data route I2S_IN-> DAC-> HP
            #
            # Left DAC to left output mixer, LINPUT3 left output mixer volume = 0dB
            # Right DAC to right output mixer, RINPUT3 right output mixer volume = 0dB
            regs[_LOUTMIX, _ROUTMIX] = 0x100
            regs[_POWER3] = 0x0C
            # Set power for DAC
            self.set_module(MODULE_DAC, True)
            self.set_module(MODULE_OMIX, True)
            self.set_module(MODULE_LINE_OUT, True)

        elif route == ROUTE_PLAYBACK_RECORD:
            #
            # Left DAC to left output mixer, LINPUT3 left output mixer volume = 0dB
            # Right DAC to right output mixer, RINPUT3 right output mixer volume = 0dB
            regs[_LOUTMIX, _ROUTMIX] = 0x100
            regs[_POWER3] = 0x3C
            self.set_module(MODULE_DAC, True)
            self.set_module(MODULE_ADC, True)
            self.set_module(MODULE_LINE_IN, True)
            self.set_module(MODULE_OMIX, True)
            self.set_module(MODULE_LINE_OUT, True)

        elif route == ROUTE_RECORD:
            # LINE_IN->ADC->I2S_OUT
            # Left and right input boost, LIN3BOOST and RIN3BOOST = 0dB
            regs[_POWER3] = 0x30
            # Power up ADC and AIN
            self.set_module(MODULE_LINE_IN, True)
            self.set_module(MODULE_ADC, True)

        else:
            raise ValueError("Invalid route")

    def set_left_input(self, input):
        if input not in self._input_config_table:
            raise ValueError("Invalid input")

        input = self._input_config_table[input]

        regs = self.regs
        if input is None:
            regs[_POWER1] = (_POWER1_AINL_MASK | _POWER1_ADCL_MASK, 0)
        elif input[0] == 0:
            regs[_POWER1] = (0, _POWER1_AINL_MASK | _POWER1_ADCL_MASK)
            regs[_INBMIX1] = input
        else:
            regs[_POWER1] = (0, _POWER1_AINL_MASK | _POWER1_ADCL_MASK | _POWER1_MICB_MASK)
            regs[_LINPATH] = input[0]
            regs[_LINVOL] = input[1]

    def set_right_input(self, input):
        if input not in self._input_config_table:
            raise ValueError("Invalid input name")

        input = self._input_config_table[input]

        regs = self.regs
        if input is None:
            regs[_POWER1] = (_POWER1_AINR_MASK | _POWER1_ADCR_MASK, 0)
        elif input[0] == 0:
            regs[_POWER1] = (0, _POWER1_AINL_MASK | _POWER1_ADCR_MASK)
            regs[_INBMIX2] = input
        else:
            regs[_POWER1] = (0, _POWER1_AINR_MASK | _POWER1_ADCR_MASK | _POWER1_MICB_MASK)
            regs[_RINPATH] = input[0]
            regs[_RINVOL] = input[1]

    def set_protocol(self, protocol):
        self.regs[_IFACE1] = (
            _IFACE1_FORMAT_MASK | _IFACE1_LRP_MASK,
            protocol,
        )

    def config_data_format(self, sysclk, sample_rate, bits):
        #  Compute sample rate divider, dac and adc are the same sample rate
        try:
            divider = self._dac_divider_table[sysclk // sample_rate]
            wl = self._audio_word_length_table[bits]
        except:
            raise ValueError("Invalid ratio sysclk/sample_rate or invalid bit length")

        self.regs[_CLOCK1] = (0x1F8, divider << 6 | divider << 3)
        self.regs[_IFACE1] = (_IFACE1_WL_MASK, wl << _IFACE1_WL_SHIFT)

    def volume(self, module, volume_l=None, volume_r=None):
        if module not in self._volume_config_table:
            raise ValueError("Invalid module")

        if volume_l is None:  # get volume
            vol_max, regnum, _ = self._volume_config_table[module]
            return (
                int((self.regs[regnum] & vol_max) * 100 / vol_max + 0.5),
                int((self.regs[regnum + 1] & vol_max) * 100 / vol_max + 0.5),
            )
        else:  # set volume
            if volume_r is None:
                volume_r = volume_l

            if not ((0 <= volume_l <= 100) and (0 <= volume_r <= 100)):
                raise ValueError("Invalid value for volume")
            elif module not in self._volume_config_table:
                raise ValueError("Invalid module")

            vol_max, regnum, flags = self._volume_config_table[module]
            self.regs[regnum] = int(volume_l * vol_max / 100 + 0.5) | flags
            self.regs[regnum + 1] = int(volume_r * vol_max / 100 + 0.5) | flags

    def mute(self, enable, soft=True, ramp=MUTE_FAST):
        enable = _DACCTL1_DACMU_MASK if enable else 0
        soft = _DACCTL2_DACSMM_MASK if soft else 0
        ramp = _DACCTL2_DACMR_MASK if ramp == MUTE_SLOW else 0
        self.regs[_DACCTL1] = (_DACCTL1_DACMU_MASK, enable)
        self.regs[_DACCTL2] = (
            _DACCTL2_DACSMM_MASK | _DACCTL2_DACMR_MASK,
            soft | ramp,
        )

    def expand_3d(self, depth=0):
        depth &= 0x0F
        cutoff = 0 if self.sample_rate >= 32000 else 0b1100000
        self.regs[_3D] = cutoff | depth << 1 | (1 if depth > 0 else 0)

    def mono(self, enable):
        enable = 1 if enable else 0
        self.regs[_DACCTL1] = (
            _DACCTL1_MONOMIX_MASK,
            enable << _DACCTL1_MONOMIX_SHIFT,
        )

    def alc_mode(self, channel, mode=ALC_MODE):
        if mode != ALC_MODE:
            mode = ALC_LIMITER
        channel &= 3
        self.regs[_ALC1] = (
            _ALC_CHANNEL_MASK,
            channel << _ALC_CHANNEL_SHIFT,
        )
        self.regs[_ALC3] = (_ALC_MODE_MASK, mode << _ALC_MODE_SHIFT)
        try:
            rate = self._alc_sample_rate_table[self.sample_rate]
        except:
            rate = 0
        self.regs[_ADDCTL3] = (_DACCTL3_ALCSR_MASK, rate)

    def alc_gain(self, target=-12, max_gain=30, min_gain=-17.25, noise_gate=-78):
        def limit(value, minval, maxval):
            value = int(value)
            if value < minval:
                value = minval
            if value > maxval:
                value = maxval
            return value

        target = limit((16 + (target * 2) // 3), 0, 15)
        max_gain = limit((max_gain + 12) // 6, 0, 7)
        min_gain = limit((min_gain * 4 + 69) // 24, 0, 7)
        noise_gate = limit((noise_gate * 2 + 153) // 3, -1, 31)
        self.regs[_ALC1] = (
            _ALC_GAIN_MASK | _ALC_TARGET_MASK,
            (max_gain << _ALC_GAIN_SHIFT) | target,
        )
        self.regs[_ALC2] = (_ALC_GAIN_MASK, (min_gain << _ALC_GAIN_SHIFT))
        if noise_gate >= 0:
            self.regs[_NOISEG] = noise_gate << _NOISEG_LEVEL_SHIFT | 1
        else:
            self.regs[_NOISEG] = 0

    def alc_time(self, attack=24, decay=192, hold=0):
        def logb(value, limit):
            value = int(value)
            lb = 0
            while value > 1:
                value >>= 1
                lb += 1
            if lb > limit:
                lb = limit
            return lb

        attack = logb(attack / 6, 7)
        decay = logb(decay / 24, 7)
        hold = logb((hold * 3) / 8, 15)
        self.regs[_ALC2] = (_ALC_HOLD_MASK, hold)
        self.regs[_ALC3] = (
            _ALC_DECAY_MASK | _ALC_ATTACK_MASK,
            (decay << _ALC_DECAY_SHIFT) | attack,
        )

    def deemphasis(self, enable):
        deem_table = (32000, 44100, 48000)
        enable = not not enable
        if enable and self.sample_rate in deem_table:
            val = deem_table.index(self.sample_rate) + 1
        else:
            val = 0
        self.regs[_DACCTL1] = (_DACCTL1_DEEM_MASK, val << _DACCTL1_DEEM_SHIFT)



# very basic test of ssl module, just to test the methods exist

try:
    import io
    import ssl
except ImportError:
    print("SKIP")
    raise SystemExit


class TestSocket(io.IOBase):
    def write(self, buf):
        return len(buf)

    def readinto(self, buf):
        return 0

    def ioctl(self, cmd, arg):
        print("TestSocket.ioctl", cmd, arg)
        return 0

    def setblocking(self, value):
        print("TestSocket.setblocking({})".format(value))


# create in client mode
try:
    ss = ssl.wrap_socket(TestSocket(), server_hostname="test.example.com")
except OSError as er:
    print("OSError: client")

# create in server mode (can use this object for further tests)
ss = ssl.wrap_socket(TestSocket(), server_side=1, do_handshake=0)

# print
print(ss)

# setblocking() propagates call to the underlying stream object
ss.setblocking(False)
ss.setblocking(True)

# write
try:
    ss.write(b"aaaa")
except OSError:
    pass

# read (underlying socket has no data)
try:
    ss.read(8)
except OSError:
    pass

# close
ss.close()
# close 2nd time
ss.close()

# read on closed socket
try:
    ss.read(10)
except OSError as er:
    print("OSError: read")

# write on closed socket
try:
    ss.write(b"aaaa")
except OSError as er:
    print("OSError: write")



# Test that socket.connect() on a socket with timeout raises EINPROGRESS or ETIMEDOUT appropriately.

import errno
import socket


def test(peer_addr, timeout, expected_exc):
    s = socket.socket()
    s.settimeout(timeout)
    try:
        s.connect(peer_addr)
        print("OK")
    except OSError as er:
        print(er.args[0] in expected_exc)
    s.close()


if __name__ == "__main__":
    # This test needs an address that doesn't respond to TCP connections.
    # 1.1.1.1:8000 seem to reliably timeout, so use that.
    addr = socket.getaddrinfo("1.1.1.1", 8000)[0][-1]

    test(addr, 0, (errno.EINPROGRESS,))
    test(addr, 1, (errno.ETIMEDOUT, "timed out"))  # CPython uses a string instead of errno



# MicroPython aioble module
# MIT license; Copyright (c) 2021 Jim Mussared

import bluetooth


log_level = 1


def log_error(*args):
    if log_level > 0:
        print("[aioble] E:", *args)


def log_warn(*args):
    if log_level > 1:
        print("[aioble] W:", *args)


def log_info(*args):
    if log_level > 2:
        print("[aioble] I:", *args)


class GattError(Exception):
    def __init__(self, status):
        self._status = status


def ensure_active():
    if not ble.active():
        try:
            from .security import load_secrets

            load_secrets()
        except:
            pass
        ble.active(True)


def config(*args, **kwargs):
    ensure_active()
    return ble.config(*args, **kwargs)


# Because different functionality is enabled by which files are available the
# different modules can register their IRQ handlers and shutdown handlers
# dynamically.
_irq_handlers = []
_shutdown_handlers = []


def register_irq_handler(irq, shutdown):
    if irq:
        _irq_handlers.append(irq)
    if shutdown:
        _shutdown_handlers.append(shutdown)


def stop():
    ble.active(False)
    for handler in _shutdown_handlers:
        handler()


# Dispatch IRQs to the registered sub-modules.
def ble_irq(event, data):
    log_info(event, data)

    for handler in _irq_handlers:
        result = handler(event, data)
        if result is not None:
            return result


# TODO: Allow this to be injected.
ble = bluetooth.BLE()
ble.irq(ble_irq)



# test machine.Signal class

try:
    import machine

    machine.PinBase
    machine.Signal
except:
    print("SKIP")
    raise SystemExit


class Pin(machine.PinBase):
    def __init__(self):
        self.v = 0

    def value(self, v=None):
        if v is None:
            return self.v
        else:
            self.v = int(v)


# test non-inverted
p = Pin()
s = machine.Signal(p)
s.value(0)
print(p.value(), s.value())
s.value(1)
print(p.value(), s.value())

# test inverted, and using on/off methods
p = Pin()
s = machine.Signal(p, invert=True)
s.off()
print(p.value(), s.value())
s.on()
print(p.value(), s.value())



# test concurrent mutating access to a shared list object
#
# MIT license; Copyright (c) 2016 Damien P. George on behalf of Pycom Ltd

import _thread

# the shared list
li = list()


# main thread function
def th(n, lo, hi):
    for repeat in range(n):
        for i in range(lo, hi):
            li.append(i)
            assert li.count(i) == repeat + 1

            li.extend([i, i])
            assert li.count(i) == repeat + 3

            li.remove(i)
            assert li.count(i) == repeat + 2

            li.remove(i)
            assert li.count(i) == repeat + 1

    with lock:
        global n_finished
        n_finished += 1


lock = _thread.allocate_lock()
n_thread = 4
n_finished = 0

# spawn threads
for i in range(n_thread):
    _thread.start_new_thread(th, (4, i * 60, (i + 1) * 60))

# busy wait for threads to finish
while n_finished < n_thread:
    pass

# check list has correct contents
li.sort()
print(li)



# test sys.path

import sys
# check that this script was executed from a file of the same name
if "__file__" not in globals() or "sys_path.py" not in __file__:
    print("SKIP")
    raise SystemExit

# test that sys.path[0] is the directory containing this script
with open(sys.path[0] + "/sys_path.py") as f:
    for _ in range(4):
        print(f.readline())



# Test that having multiple threads block on stdin doesn't cause any issues.
#
# The test doesn't expect any input on stdin.
#
# This is a regression test for https://github.com/micropython/micropython/issues/15230
# on rp2, but doubles as a general property to test across all ports.
import sys
import _thread

try:
    import select
except ImportError:
    print("SKIP")
    raise SystemExit


class StdinWaiter:
    def __init__(self):
        self._done = False

    def wait_stdin(self, timeout_ms):
        poller = select.poll()
        poller.register(sys.stdin, select.POLLIN)
        poller.poll(timeout_ms)
        # Ignoring the poll result as we don't expect any input
        self._done = True

    def is_done(self):
        return self._done


thread_waiter = StdinWaiter()
_thread.start_new_thread(thread_waiter.wait_stdin, (1000,))
StdinWaiter().wait_stdin(1000)

# Spinning here is mostly not necessary but there is some inconsistency waking
# the two threads, especially on CPython CI runners where the thread may not
# have run yet. The actual delay is <20ms but spinning here instead of
# sleep(0.1) means the test can run on MP builds without float support.
while not thread_waiter.is_done():
    pass

# The background thread should have completed its wait by now.
print(thread_waiter.is_done())



#
# definitions
#
MACHINE_RA4M1_CLICKER = "RA4M1 CLICKER with RA4M1"
MACHINE_EK_RA4M1 = "EK-RA4M1 with RA4M1"
MACHINE_EK_RA4W1 = "EK-RA4W1 with RA4W1"
MACHINE_EK_RA6M1 = "EK-RA6M1 with RA6M1"
MACHINE_EK_RA6M2 = "EK-RA6M2 with RA6M2"
SYSCLK_RA4M1_CLICKER = 48000000
SYSCLK_EK_RA4M1 = 48000000
SYSCLK_EK_RA4W1 = 48000000
SYSCLK_EK_RA6M1 = 120000000
SYSCLK_EK_RA6M2 = 120000000

#
# machine
#

import os

try:
    import machine
except ImportError:
    print("machine module is not found")
    raise SystemExit

m = os.uname().machine
f = machine.freq()

if m == MACHINE_RA4M1_CLICKER:
    if f == SYSCLK_RA4M1_CLICKER:
        print("freq: OK")
    else:
        print("freq: NG")


if m == MACHINE_EK_RA4M1:
    if f == SYSCLK_EK_RA4M1:
        print("freq: OK")
    else:
        print("freq: NG")

if m == MACHINE_EK_RA4W1:
    if f == SYSCLK_EK_RA4W1:
        print("freq: OK")
    else:
        print("freq: NG")

if m == MACHINE_EK_RA6M1:
    if f == SYSCLK_EK_RA6M1:
        print("freq: OK")
    else:
        print("freq: NG")

if m == MACHINE_EK_RA6M2:
    if f == SYSCLK_EK_RA6M2:
        print("freq: OK")
    else:
        print("freq: NG")



import vfs
import machine, rp2


# Try to mount the filesystem, and format the flash if it doesn't exist.
bdev = rp2.Flash()
try:
    vfs.mount(vfs.VfsFat(bdev), "/")
except:
    vfs.VfsFat.mkfs(bdev)
    vfs.mount(vfs.VfsFat(bdev), "/")

del vfs, bdev



#!/usr/bin/env python

r"""
Flash the WiPy (format, update service pack and program).

Example:

> python uniflash.py -u "C:\ti\uniflash_3.2\uniflashCLI.bat" -c "C:\VirtualBoxShared\GitHub\wipy_uniflash.usf" -p 8 -s "C:\ti\CC31xx_CC32xx_ServicePack_1.0.0.10.0\servicepack_1.0.0.10.0.bin"

or:

> python uniflash.py -u "C:\ti\uniflash_3.2\uniflashCLI.bat" -c "C:\VirtualBoxShared\GitHub\launchxl_uniflash.usf" -p 8 -s "C:\ti\CC31xx_CC32xx_ServicePack_1.0.0.10.0\servicepack_1.0.0.10.0.bin"

"""

import sys
import argparse
import subprocess


def print_exception(e):
    print("Exception: {}, on line {}".format(e, sys.exc_info()[-1].tb_lineno))


def execute(command):
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    cmd_log = ""

    # Poll process for new output until finished
    while True:
        nextline = process.stdout.readline()
        if nextline == "" and process.poll() != None:
            break
        sys.stdout.write(nextline)
        sys.stdout.flush()
        cmd_log += nextline

    output = process.communicate()[0]
    exitCode = process.returncode

    if exitCode == 0:
        return cmd_log
    else:
        raise subprocess.CalledProcessError(exitCode, command, output)


def main():
    cmd_parser = argparse.ArgumentParser(
        description="Flash the WiPy and optionally run a small test on it."
    )
    cmd_parser.add_argument(
        "-u", "--uniflash", default=None, help="the path to the uniflash cli executable"
    )
    cmd_parser.add_argument(
        "-c", "--config", default=None, help="the path to the uniflash config file"
    )
    cmd_parser.add_argument("-p", "--port", default=8, help="the com serial port")
    cmd_parser.add_argument(
        "-s", "--servicepack", default=None, help="the path to the servicepack file"
    )
    args = cmd_parser.parse_args()

    output = ""
    com_port = "com=" + str(args.port)
    servicepack_path = "spPath=" + args.servicepack

    try:
        if args.uniflash == None or args.config == None:
            raise ValueError("uniflash path and config path are mandatory")
        if args.servicepack == None:
            output += execute(
                [
                    args.uniflash,
                    "-config",
                    args.config,
                    "-setOptions",
                    com_port,
                    "-operations",
                    "format",
                    "program",
                ]
            )
        else:
            output += execute(
                [
                    args.uniflash,
                    "-config",
                    args.config,
                    "-setOptions",
                    com_port,
                    servicepack_path,
                    "-operations",
                    "format",
                    "servicePackUpdate",
                    "program",
                ]
            )
    except Exception as e:
        print_exception(e)
        output = ""
    finally:
        if "Finish Executing operation: program" in output:
            print("======================================")
            print("Board programmed OK")
            print("======================================")
            sys.exit(0)
        else:
            print("======================================")
            print("ERROR: Programming failed!")
            print("======================================")
            sys.exit(1)


if __name__ == "__main__":
    main()



# Retrieve the native architecture of the target.
# See https://docs.micropython.org/en/latest/reference/mpyfiles.html#versioning-and-compatibility-of-mpy-files
# for more details.

import sys

sys_mpy = getattr(sys.implementation, "_mpy", 0)
arch = [
    None,
    "x86",
    "x64",
    "armv6",
    "armv6m",
    "armv7m",
    "armv7em",
    "armv7emsp",
    "armv7emdp",
    "xtensa",
    "xtensawin",
    "rv32imc",
][sys_mpy >> 10]
print(arch)



#! /usr/bin/env python3

from test import support
import unittest
import urllib.parse

RFC1808_BASE = "http://a/b/c/d;p?q#f"
RFC2396_BASE = "http://a/b/c/d;p?q"
RFC3986_BASE = "http://a/b/c/d;p?q"
SIMPLE_BASE = "http://a/b/c/d"

# A list of test cases.  Each test case is a two-tuple that contains
# a string with the query and a dictionary with the expected result.

parse_qsl_test_cases = [
    ("", []),
    ("&", []),
    ("&&", []),
    ("=", [("", "")]),
    ("=a", [("", "a")]),
    ("a", [("a", "")]),
    ("a=", [("a", "")]),
    ("a=", [("a", "")]),
    ("&a=b", [("a", "b")]),
    ("a=a+b&b=b+c", [("a", "a b"), ("b", "b c")]),
    ("a=1&a=2", [("a", "1"), ("a", "2")]),
    (b"", []),
    (b"&", []),
    (b"&&", []),
    (b"=", [(b"", b"")]),
    (b"=a", [(b"", b"a")]),
    (b"a", [(b"a", b"")]),
    (b"a=", [(b"a", b"")]),
    (b"a=", [(b"a", b"")]),
    (b"&a=b", [(b"a", b"b")]),
    (b"a=a+b&b=b+c", [(b"a", b"a b"), (b"b", b"b c")]),
    (b"a=1&a=2", [(b"a", b"1"), (b"a", b"2")]),
]


class UrlParseTestCase(unittest.TestCase):
    def checkRoundtrips(self, url, parsed, split):
        result = urllib.parse.urlparse(url)
        self.assertEqual(result, parsed)
        t = (
            result.scheme,
            result.netloc,
            result.path,
            result.params,
            result.query,
            result.fragment,
        )
        self.assertEqual(t, parsed)
        # put it back together and it should be the same
        result2 = urllib.parse.urlunparse(result)
        self.assertEqual(result2, url)
        self.assertEqual(result2, result.geturl())

        # the result of geturl() is a fixpoint; we can always parse it
        # again to get the same result:
        result3 = urllib.parse.urlparse(result.geturl())
        self.assertEqual(result3.geturl(), result.geturl())
        self.assertEqual(result3, result)
        self.assertEqual(result3.scheme, result.scheme)
        self.assertEqual(result3.netloc, result.netloc)
        self.assertEqual(result3.path, result.path)
        self.assertEqual(result3.params, result.params)
        self.assertEqual(result3.query, result.query)
        self.assertEqual(result3.fragment, result.fragment)
        self.assertEqual(result3.username, result.username)
        self.assertEqual(result3.password, result.password)
        self.assertEqual(result3.hostname, result.hostname)
        self.assertEqual(result3.port, result.port)

        # check the roundtrip using urlsplit() as well
        result = urllib.parse.urlsplit(url)
        self.assertEqual(result, split)
        t = (result.scheme, result.netloc, result.path, result.query, result.fragment)
        self.assertEqual(t, split)
        result2 = urllib.parse.urlunsplit(result)
        self.assertEqual(result2, url)
        self.assertEqual(result2, result.geturl())

        # check the fixpoint property of re-parsing the result of geturl()
        result3 = urllib.parse.urlsplit(result.geturl())
        self.assertEqual(result3.geturl(), result.geturl())
        self.assertEqual(result3, result)
        self.assertEqual(result3.scheme, result.scheme)
        self.assertEqual(result3.netloc, result.netloc)
        self.assertEqual(result3.path, result.path)
        self.assertEqual(result3.query, result.query)
        self.assertEqual(result3.fragment, result.fragment)
        self.assertEqual(result3.username, result.username)
        self.assertEqual(result3.password, result.password)
        self.assertEqual(result3.hostname, result.hostname)
        self.assertEqual(result3.port, result.port)

    def test_qsl(self):
        for orig, expect in parse_qsl_test_cases:
            result = urllib.parse.parse_qsl(orig, keep_blank_values=True)
            self.assertEqual(result, expect, "Error parsing %r" % orig)
            expect_without_blanks = [v for v in expect if len(v[1])]
            result = urllib.parse.parse_qsl(orig, keep_blank_values=False)
            self.assertEqual(result, expect_without_blanks, "Error parsing %r" % orig)

    def test_roundtrips(self):
        str_cases = [
            (
                "file:///tmp/junk.txt",
                ("file", "", "/tmp/junk.txt", "", "", ""),
                ("file", "", "/tmp/junk.txt", "", ""),
            ),
            (
                "imap://mail.python.org/mbox1",
                ("imap", "mail.python.org", "/mbox1", "", "", ""),
                ("imap", "mail.python.org", "/mbox1", "", ""),
            ),
            (
                "mms://wms.sys.hinet.net/cts/Drama/09006251100.asf",
                ("mms", "wms.sys.hinet.net", "/cts/Drama/09006251100.asf", "", "", ""),
                ("mms", "wms.sys.hinet.net", "/cts/Drama/09006251100.asf", "", ""),
            ),
            (
                "nfs://server/path/to/file.txt",
                ("nfs", "server", "/path/to/file.txt", "", "", ""),
                ("nfs", "server", "/path/to/file.txt", "", ""),
            ),
            (
                "svn+ssh://svn.zope.org/repos/main/ZConfig/trunk/",
                ("svn+ssh", "svn.zope.org", "/repos/main/ZConfig/trunk/", "", "", ""),
                ("svn+ssh", "svn.zope.org", "/repos/main/ZConfig/trunk/", "", ""),
            ),
            (
                "git+ssh://git@github.com/user/project.git",
                ("git+ssh", "git@github.com", "/user/project.git", "", "", ""),
                ("git+ssh", "git@github.com", "/user/project.git", "", ""),
            ),
        ]

        def _encode(t):
            return (
                t[0].encode("ascii"),
                tuple(x.encode("ascii") for x in t[1]),
                tuple(x.encode("ascii") for x in t[2]),
            )

        bytes_cases = [_encode(x) for x in str_cases]
        for url, parsed, split in str_cases + bytes_cases:
            self.checkRoundtrips(url, parsed, split)

    def test_http_roundtrips(self):
        # urllib.parse.urlsplit treats 'http:' as an optimized special case,
        # so we test both 'http:' and 'https:' in all the following.
        # Three cheers for white box knowledge!
        str_cases = [
            (
                "://www.python.org",
                ("www.python.org", "", "", "", ""),
                ("www.python.org", "", "", ""),
            ),
            (
                "://www.python.org#abc",
                ("www.python.org", "", "", "", "abc"),
                ("www.python.org", "", "", "abc"),
            ),
            (
                "://www.python.org?q=abc",
                ("www.python.org", "", "", "q=abc", ""),
                ("www.python.org", "", "q=abc", ""),
            ),
            (
                "://www.python.org/#abc",
                ("www.python.org", "/", "", "", "abc"),
                ("www.python.org", "/", "", "abc"),
            ),
            ("://a/b/c/d;p?q#f", ("a", "/b/c/d", "p", "q", "f"), ("a", "/b/c/d;p", "q", "f")),
        ]

        def _encode(t):
            return (
                t[0].encode("ascii"),
                tuple(x.encode("ascii") for x in t[1]),
                tuple(x.encode("ascii") for x in t[2]),
            )

        bytes_cases = [_encode(x) for x in str_cases]
        str_schemes = ("http", "https")
        bytes_schemes = (b"http", b"https")
        str_tests = str_schemes, str_cases
        bytes_tests = bytes_schemes, bytes_cases
        for schemes, test_cases in (str_tests, bytes_tests):
            for scheme in schemes:
                for url, parsed, split in test_cases:
                    url = scheme + url
                    parsed = (scheme,) + parsed
                    split = (scheme,) + split
                    self.checkRoundtrips(url, parsed, split)

    def checkJoin(self, base, relurl, expected):
        str_components = (base, relurl, expected)
        self.assertEqual(urllib.parse.urljoin(base, relurl), expected)
        bytes_components = baseb, relurlb, expectedb = [x.encode("ascii") for x in str_components]
        self.assertEqual(urllib.parse.urljoin(baseb, relurlb), expectedb)

    def test_unparse_parse(self):
        str_cases = [
            "Python",
            "./Python",
            "x-newscheme://foo.com/stuff",
            "x://y",
            "x:/y",
            "x:/",
            "/",
        ]
        bytes_cases = [x.encode("ascii") for x in str_cases]
        for u in str_cases + bytes_cases:
            self.assertEqual(urllib.parse.urlunsplit(urllib.parse.urlsplit(u)), u)
            self.assertEqual(urllib.parse.urlunparse(urllib.parse.urlparse(u)), u)

    def test_RFC1808(self):
        # "normal" cases from RFC 1808:
        self.checkJoin(RFC1808_BASE, "g:h", "g:h")
        self.checkJoin(RFC1808_BASE, "g", "http://a/b/c/g")
        self.checkJoin(RFC1808_BASE, "./g", "http://a/b/c/g")
        self.checkJoin(RFC1808_BASE, "g/", "http://a/b/c/g/")
        self.checkJoin(RFC1808_BASE, "/g", "http://a/g")
        self.checkJoin(RFC1808_BASE, "//g", "http://g")
        self.checkJoin(RFC1808_BASE, "g?y", "http://a/b/c/g?y")
        self.checkJoin(RFC1808_BASE, "g?y/./x", "http://a/b/c/g?y/./x")
        self.checkJoin(RFC1808_BASE, "#s", "http://a/b/c/d;p?q#s")
        self.checkJoin(RFC1808_BASE, "g#s", "http://a/b/c/g#s")
        self.checkJoin(RFC1808_BASE, "g#s/./x", "http://a/b/c/g#s/./x")
        self.checkJoin(RFC1808_BASE, "g?y#s", "http://a/b/c/g?y#s")
        self.checkJoin(RFC1808_BASE, "g;x", "http://a/b/c/g;x")
        self.checkJoin(RFC1808_BASE, "g;x?y#s", "http://a/b/c/g;x?y#s")
        self.checkJoin(RFC1808_BASE, ".", "http://a/b/c/")
        self.checkJoin(RFC1808_BASE, "./", "http://a/b/c/")
        self.checkJoin(RFC1808_BASE, "..", "http://a/b/")
        self.checkJoin(RFC1808_BASE, "../", "http://a/b/")
        self.checkJoin(RFC1808_BASE, "../g", "http://a/b/g")
        self.checkJoin(RFC1808_BASE, "../..", "http://a/")
        self.checkJoin(RFC1808_BASE, "../../", "http://a/")
        self.checkJoin(RFC1808_BASE, "../../g", "http://a/g")

        # "abnormal" cases from RFC 1808:
        self.checkJoin(RFC1808_BASE, "", "http://a/b/c/d;p?q#f")
        self.checkJoin(RFC1808_BASE, "../../../g", "http://a/../g")
        self.checkJoin(RFC1808_BASE, "../../../../g", "http://a/../../g")
        self.checkJoin(RFC1808_BASE, "/./g", "http://a/./g")
        self.checkJoin(RFC1808_BASE, "/../g", "http://a/../g")
        self.checkJoin(RFC1808_BASE, "g.", "http://a/b/c/g.")
        self.checkJoin(RFC1808_BASE, ".g", "http://a/b/c/.g")
        self.checkJoin(RFC1808_BASE, "g..", "http://a/b/c/g..")
        self.checkJoin(RFC1808_BASE, "..g", "http://a/b/c/..g")
        self.checkJoin(RFC1808_BASE, "./../g", "http://a/b/g")
        self.checkJoin(RFC1808_BASE, "./g/.", "http://a/b/c/g/")
        self.checkJoin(RFC1808_BASE, "g/./h", "http://a/b/c/g/h")
        self.checkJoin(RFC1808_BASE, "g/../h", "http://a/b/c/h")

        # RFC 1808 and RFC 1630 disagree on these (according to RFC 1808),
        # so we'll not actually run these tests (which expect 1808 behavior).
        # self.checkJoin(RFC1808_BASE, 'http:g', 'http:g')
        # self.checkJoin(RFC1808_BASE, 'http:', 'http:')

    def test_RFC2368(self):
        # Issue 11467: path that starts with a number is not parsed correctly
        self.assertEqual(
            urllib.parse.urlparse("mailto:1337@example.org"),
            ("mailto", "", "1337@example.org", "", "", ""),
        )

    def test_RFC2396(self):
        # cases from RFC 2396

        self.checkJoin(RFC2396_BASE, "g:h", "g:h")
        self.checkJoin(RFC2396_BASE, "g", "http://a/b/c/g")
        self.checkJoin(RFC2396_BASE, "./g", "http://a/b/c/g")
        self.checkJoin(RFC2396_BASE, "g/", "http://a/b/c/g/")
        self.checkJoin(RFC2396_BASE, "/g", "http://a/g")
        self.checkJoin(RFC2396_BASE, "//g", "http://g")
        self.checkJoin(RFC2396_BASE, "g?y", "http://a/b/c/g?y")
        self.checkJoin(RFC2396_BASE, "#s", "http://a/b/c/d;p?q#s")
        self.checkJoin(RFC2396_BASE, "g#s", "http://a/b/c/g#s")
        self.checkJoin(RFC2396_BASE, "g?y#s", "http://a/b/c/g?y#s")
        self.checkJoin(RFC2396_BASE, "g;x", "http://a/b/c/g;x")
        self.checkJoin(RFC2396_BASE, "g;x?y#s", "http://a/b/c/g;x?y#s")
        self.checkJoin(RFC2396_BASE, ".", "http://a/b/c/")
        self.checkJoin(RFC2396_BASE, "./", "http://a/b/c/")
        self.checkJoin(RFC2396_BASE, "..", "http://a/b/")
        self.checkJoin(RFC2396_BASE, "../", "http://a/b/")
        self.checkJoin(RFC2396_BASE, "../g", "http://a/b/g")
        self.checkJoin(RFC2396_BASE, "../..", "http://a/")
        self.checkJoin(RFC2396_BASE, "../../", "http://a/")
        self.checkJoin(RFC2396_BASE, "../../g", "http://a/g")
        self.checkJoin(RFC2396_BASE, "", RFC2396_BASE)
        self.checkJoin(RFC2396_BASE, "../../../g", "http://a/../g")
        self.checkJoin(RFC2396_BASE, "../../../../g", "http://a/../../g")
        self.checkJoin(RFC2396_BASE, "/./g", "http://a/./g")
        self.checkJoin(RFC2396_BASE, "/../g", "http://a/../g")
        self.checkJoin(RFC2396_BASE, "g.", "http://a/b/c/g.")
        self.checkJoin(RFC2396_BASE, ".g", "http://a/b/c/.g")
        self.checkJoin(RFC2396_BASE, "g..", "http://a/b/c/g..")
        self.checkJoin(RFC2396_BASE, "..g", "http://a/b/c/..g")
        self.checkJoin(RFC2396_BASE, "./../g", "http://a/b/g")
        self.checkJoin(RFC2396_BASE, "./g/.", "http://a/b/c/g/")
        self.checkJoin(RFC2396_BASE, "g/./h", "http://a/b/c/g/h")
        self.checkJoin(RFC2396_BASE, "g/../h", "http://a/b/c/h")
        self.checkJoin(RFC2396_BASE, "g;x=1/./y", "http://a/b/c/g;x=1/y")
        self.checkJoin(RFC2396_BASE, "g;x=1/../y", "http://a/b/c/y")
        self.checkJoin(RFC2396_BASE, "g?y/./x", "http://a/b/c/g?y/./x")
        self.checkJoin(RFC2396_BASE, "g?y/../x", "http://a/b/c/g?y/../x")
        self.checkJoin(RFC2396_BASE, "g#s/./x", "http://a/b/c/g#s/./x")
        self.checkJoin(RFC2396_BASE, "g#s/../x", "http://a/b/c/g#s/../x")

    def test_RFC3986(self):
        # Test cases from RFC3986
        self.checkJoin(RFC3986_BASE, "?y", "http://a/b/c/d;p?y")
        self.checkJoin(RFC2396_BASE, ";x", "http://a/b/c/;x")
        self.checkJoin(RFC3986_BASE, "g:h", "g:h")
        self.checkJoin(RFC3986_BASE, "g", "http://a/b/c/g")
        self.checkJoin(RFC3986_BASE, "./g", "http://a/b/c/g")
        self.checkJoin(RFC3986_BASE, "g/", "http://a/b/c/g/")
        self.checkJoin(RFC3986_BASE, "/g", "http://a/g")
        self.checkJoin(RFC3986_BASE, "//g", "http://g")
        self.checkJoin(RFC3986_BASE, "?y", "http://a/b/c/d;p?y")
        self.checkJoin(RFC3986_BASE, "g?y", "http://a/b/c/g?y")
        self.checkJoin(RFC3986_BASE, "#s", "http://a/b/c/d;p?q#s")
        self.checkJoin(RFC3986_BASE, "g#s", "http://a/b/c/g#s")
        self.checkJoin(RFC3986_BASE, "g?y#s", "http://a/b/c/g?y#s")
        self.checkJoin(RFC3986_BASE, ";x", "http://a/b/c/;x")
        self.checkJoin(RFC3986_BASE, "g;x", "http://a/b/c/g;x")
        self.checkJoin(RFC3986_BASE, "g;x?y#s", "http://a/b/c/g;x?y#s")
        self.checkJoin(RFC3986_BASE, "", "http://a/b/c/d;p?q")
        self.checkJoin(RFC3986_BASE, ".", "http://a/b/c/")
        self.checkJoin(RFC3986_BASE, "./", "http://a/b/c/")
        self.checkJoin(RFC3986_BASE, "..", "http://a/b/")
        self.checkJoin(RFC3986_BASE, "../", "http://a/b/")
        self.checkJoin(RFC3986_BASE, "../g", "http://a/b/g")
        self.checkJoin(RFC3986_BASE, "../..", "http://a/")
        self.checkJoin(RFC3986_BASE, "../../", "http://a/")
        self.checkJoin(RFC3986_BASE, "../../g", "http://a/g")

        # Abnormal Examples

        # The 'abnormal scenarios' are incompatible with RFC2986 parsing
        # Tests are here for reference.

        # self.checkJoin(RFC3986_BASE, '../../../g','http://a/g')
        # self.checkJoin(RFC3986_BASE, '../../../../g','http://a/g')
        # self.checkJoin(RFC3986_BASE, '/./g','http://a/g')
        # self.checkJoin(RFC3986_BASE, '/../g','http://a/g')

        self.checkJoin(RFC3986_BASE, "g.", "http://a/b/c/g.")
        self.checkJoin(RFC3986_BASE, ".g", "http://a/b/c/.g")
        self.checkJoin(RFC3986_BASE, "g..", "http://a/b/c/g..")
        self.checkJoin(RFC3986_BASE, "..g", "http://a/b/c/..g")
        self.checkJoin(RFC3986_BASE, "./../g", "http://a/b/g")
        self.checkJoin(RFC3986_BASE, "./g/.", "http://a/b/c/g/")
        self.checkJoin(RFC3986_BASE, "g/./h", "http://a/b/c/g/h")
        self.checkJoin(RFC3986_BASE, "g/../h", "http://a/b/c/h")
        self.checkJoin(RFC3986_BASE, "g;x=1/./y", "http://a/b/c/g;x=1/y")
        self.checkJoin(RFC3986_BASE, "g;x=1/../y", "http://a/b/c/y")
        self.checkJoin(RFC3986_BASE, "g?y/./x", "http://a/b/c/g?y/./x")
        self.checkJoin(RFC3986_BASE, "g?y/../x", "http://a/b/c/g?y/../x")
        self.checkJoin(RFC3986_BASE, "g#s/./x", "http://a/b/c/g#s/./x")
        self.checkJoin(RFC3986_BASE, "g#s/../x", "http://a/b/c/g#s/../x")
        # self.checkJoin(RFC3986_BASE, 'http:g','http:g') # strict parser
        self.checkJoin(RFC3986_BASE, "http:g", "http://a/b/c/g")  # relaxed parser

        # Test for issue9721
        self.checkJoin("http://a/b/c/de", ";x", "http://a/b/c/;x")

    def test_urljoins(self):
        self.checkJoin(SIMPLE_BASE, "g:h", "g:h")
        self.checkJoin(SIMPLE_BASE, "http:g", "http://a/b/c/g")
        self.checkJoin(SIMPLE_BASE, "http:", "http://a/b/c/d")
        self.checkJoin(SIMPLE_BASE, "g", "http://a/b/c/g")
        self.checkJoin(SIMPLE_BASE, "./g", "http://a/b/c/g")
        self.checkJoin(SIMPLE_BASE, "g/", "http://a/b/c/g/")
        self.checkJoin(SIMPLE_BASE, "/g", "http://a/g")
        self.checkJoin(SIMPLE_BASE, "//g", "http://g")
        self.checkJoin(SIMPLE_BASE, "?y", "http://a/b/c/d?y")
        self.checkJoin(SIMPLE_BASE, "g?y", "http://a/b/c/g?y")
        self.checkJoin(SIMPLE_BASE, "g?y/./x", "http://a/b/c/g?y/./x")
        self.checkJoin(SIMPLE_BASE, ".", "http://a/b/c/")
        self.checkJoin(SIMPLE_BASE, "./", "http://a/b/c/")
        self.checkJoin(SIMPLE_BASE, "..", "http://a/b/")
        self.checkJoin(SIMPLE_BASE, "../", "http://a/b/")
        self.checkJoin(SIMPLE_BASE, "../g", "http://a/b/g")
        self.checkJoin(SIMPLE_BASE, "../..", "http://a/")
        self.checkJoin(SIMPLE_BASE, "../../g", "http://a/g")
        self.checkJoin(SIMPLE_BASE, "../../../g", "http://a/../g")
        self.checkJoin(SIMPLE_BASE, "./../g", "http://a/b/g")
        self.checkJoin(SIMPLE_BASE, "./g/.", "http://a/b/c/g/")
        self.checkJoin(SIMPLE_BASE, "/./g", "http://a/./g")
        self.checkJoin(SIMPLE_BASE, "g/./h", "http://a/b/c/g/h")
        self.checkJoin(SIMPLE_BASE, "g/../h", "http://a/b/c/h")
        self.checkJoin(SIMPLE_BASE, "http:g", "http://a/b/c/g")
        self.checkJoin(SIMPLE_BASE, "http:", "http://a/b/c/d")
        self.checkJoin(SIMPLE_BASE, "http:?y", "http://a/b/c/d?y")
        self.checkJoin(SIMPLE_BASE, "http:g?y", "http://a/b/c/g?y")
        self.checkJoin(SIMPLE_BASE, "http:g?y/./x", "http://a/b/c/g?y/./x")
        self.checkJoin("http:///", "..", "http:///")
        self.checkJoin("", "http://a/b/c/g?y/./x", "http://a/b/c/g?y/./x")
        self.checkJoin("", "http://a/./g", "http://a/./g")
        self.checkJoin("svn://pathtorepo/dir1", "dir2", "svn://pathtorepo/dir2")
        self.checkJoin("svn+ssh://pathtorepo/dir1", "dir2", "svn+ssh://pathtorepo/dir2")

    def test_RFC2732(self):
        str_cases = [
            ("http://Test.python.org:5432/foo/", "test.python.org", 5432),
            ("http://12.34.56.78:5432/foo/", "12.34.56.78", 5432),
            ("http://[::1]:5432/foo/", "::1", 5432),
            ("http://[dead:beef::1]:5432/foo/", "dead:beef::1", 5432),
            ("http://[dead:beef::]:5432/foo/", "dead:beef::", 5432),
            (
                "http://[dead:beef:cafe:5417:affe:8FA3:deaf:feed]:5432/foo/",
                "dead:beef:cafe:5417:affe:8fa3:deaf:feed",
                5432,
            ),
            ("http://[::12.34.56.78]:5432/foo/", "::12.34.56.78", 5432),
            ("http://[::ffff:12.34.56.78]:5432/foo/", "::ffff:12.34.56.78", 5432),
            ("http://Test.python.org/foo/", "test.python.org", None),
            ("http://12.34.56.78/foo/", "12.34.56.78", None),
            ("http://[::1]/foo/", "::1", None),
            ("http://[dead:beef::1]/foo/", "dead:beef::1", None),
            ("http://[dead:beef::]/foo/", "dead:beef::", None),
            (
                "http://[dead:beef:cafe:5417:affe:8FA3:deaf:feed]/foo/",
                "dead:beef:cafe:5417:affe:8fa3:deaf:feed",
                None,
            ),
            ("http://[::12.34.56.78]/foo/", "::12.34.56.78", None),
            ("http://[::ffff:12.34.56.78]/foo/", "::ffff:12.34.56.78", None),
        ]

        def _encode(t):
            return t[0].encode("ascii"), t[1].encode("ascii"), t[2]

        bytes_cases = [_encode(x) for x in str_cases]
        for url, hostname, port in str_cases + bytes_cases:
            urlparsed = urllib.parse.urlparse(url)
            self.assertEqual((urlparsed.hostname, urlparsed.port), (hostname, port))

        str_cases = [
            "http://::12.34.56.78]/",
            "http://[::1/foo/",
            "ftp://[::1/foo/bad]/bad",
            "http://[::1/foo/bad]/bad",
            "http://[::ffff:12.34.56.78",
        ]
        bytes_cases = [x.encode("ascii") for x in str_cases]
        for invalid_url in str_cases + bytes_cases:
            self.assertRaises(ValueError, urllib.parse.urlparse, invalid_url)

    def test_urldefrag(self):
        str_cases = [
            ("http://python.org#frag", "http://python.org", "frag"),
            ("http://python.org", "http://python.org", ""),
            ("http://python.org/#frag", "http://python.org/", "frag"),
            ("http://python.org/", "http://python.org/", ""),
            ("http://python.org/?q#frag", "http://python.org/?q", "frag"),
            ("http://python.org/?q", "http://python.org/?q", ""),
            ("http://python.org/p#frag", "http://python.org/p", "frag"),
            ("http://python.org/p?q", "http://python.org/p?q", ""),
            (RFC1808_BASE, "http://a/b/c/d;p?q", "f"),
            (RFC2396_BASE, "http://a/b/c/d;p?q", ""),
        ]

        def _encode(t):
            return type(t)(x.encode("ascii") for x in t)

        bytes_cases = [_encode(x) for x in str_cases]
        for url, defrag, frag in str_cases + bytes_cases:
            result = urllib.parse.urldefrag(url)
            self.assertEqual(result.geturl(), url)
            self.assertEqual(result, (defrag, frag))
            self.assertEqual(result.url, defrag)
            self.assertEqual(result.fragment, frag)

    def test_urlsplit_attributes(self):
        url = "HTTP://WWW.PYTHON.ORG/doc/#frag"
        p = urllib.parse.urlsplit(url)
        self.assertEqual(p.scheme, "http")
        self.assertEqual(p.netloc, "WWW.PYTHON.ORG")
        self.assertEqual(p.path, "/doc/")
        self.assertEqual(p.query, "")
        self.assertEqual(p.fragment, "frag")
        self.assertEqual(p.username, None)
        self.assertEqual(p.password, None)
        self.assertEqual(p.hostname, "www.python.org")
        self.assertEqual(p.port, None)
        # geturl() won't return exactly the original URL in this case
        # since the scheme is always case-normalized
        # We handle this by ignoring the first 4 characters of the URL
        self.assertEqual(p.geturl()[4:], url[4:])

        url = "http://User:Pass@www.python.org:080/doc/?query=yes#frag"
        p = urllib.parse.urlsplit(url)
        self.assertEqual(p.scheme, "http")
        self.assertEqual(p.netloc, "User:Pass@www.python.org:080")
        self.assertEqual(p.path, "/doc/")
        self.assertEqual(p.query, "query=yes")
        self.assertEqual(p.fragment, "frag")
        self.assertEqual(p.username, "User")
        self.assertEqual(p.password, "Pass")
        self.assertEqual(p.hostname, "www.python.org")
        self.assertEqual(p.port, 80)
        self.assertEqual(p.geturl(), url)

        # Addressing issue1698, which suggests Username can contain
        # "@" characters.  Though not RFC compliant, many ftp sites allow
        # and request email addresses as usernames.

        url = "http://User@example.com:Pass@www.python.org:080/doc/?query=yes#frag"
        p = urllib.parse.urlsplit(url)
        self.assertEqual(p.scheme, "http")
        self.assertEqual(p.netloc, "User@example.com:Pass@www.python.org:080")
        self.assertEqual(p.path, "/doc/")
        self.assertEqual(p.query, "query=yes")
        self.assertEqual(p.fragment, "frag")
        self.assertEqual(p.username, "User@example.com")
        self.assertEqual(p.password, "Pass")
        self.assertEqual(p.hostname, "www.python.org")
        self.assertEqual(p.port, 80)
        self.assertEqual(p.geturl(), url)

        # And check them all again, only with bytes this time
        url = b"HTTP://WWW.PYTHON.ORG/doc/#frag"
        p = urllib.parse.urlsplit(url)
        self.assertEqual(p.scheme, b"http")
        self.assertEqual(p.netloc, b"WWW.PYTHON.ORG")
        self.assertEqual(p.path, b"/doc/")
        self.assertEqual(p.query, b"")
        self.assertEqual(p.fragment, b"frag")
        self.assertEqual(p.username, None)
        self.assertEqual(p.password, None)
        self.assertEqual(p.hostname, b"www.python.org")
        self.assertEqual(p.port, None)
        self.assertEqual(p.geturl()[4:], url[4:])

        url = b"http://User:Pass@www.python.org:080/doc/?query=yes#frag"
        p = urllib.parse.urlsplit(url)
        self.assertEqual(p.scheme, b"http")
        self.assertEqual(p.netloc, b"User:Pass@www.python.org:080")
        self.assertEqual(p.path, b"/doc/")
        self.assertEqual(p.query, b"query=yes")
        self.assertEqual(p.fragment, b"frag")
        self.assertEqual(p.username, b"User")
        self.assertEqual(p.password, b"Pass")
        self.assertEqual(p.hostname, b"www.python.org")
        self.assertEqual(p.port, 80)
        self.assertEqual(p.geturl(), url)

        url = b"http://User@example.com:Pass@www.python.org:080/doc/?query=yes#frag"
        p = urllib.parse.urlsplit(url)
        self.assertEqual(p.scheme, b"http")
        self.assertEqual(p.netloc, b"User@example.com:Pass@www.python.org:080")
        self.assertEqual(p.path, b"/doc/")
        self.assertEqual(p.query, b"query=yes")
        self.assertEqual(p.fragment, b"frag")
        self.assertEqual(p.username, b"User@example.com")
        self.assertEqual(p.password, b"Pass")
        self.assertEqual(p.hostname, b"www.python.org")
        self.assertEqual(p.port, 80)
        self.assertEqual(p.geturl(), url)

        # Verify an illegal port is returned as None
        url = b"HTTP://WWW.PYTHON.ORG:65536/doc/#frag"
        p = urllib.parse.urlsplit(url)
        self.assertEqual(p.port, None)

    def test_attributes_bad_port(self):
        """Check handling of non-integer ports."""
        p = urllib.parse.urlsplit("http://www.example.net:foo")
        self.assertEqual(p.netloc, "www.example.net:foo")
        self.assertRaises(ValueError, lambda: p.port)

        p = urllib.parse.urlparse("http://www.example.net:foo")
        self.assertEqual(p.netloc, "www.example.net:foo")
        self.assertRaises(ValueError, lambda: p.port)

        # Once again, repeat ourselves to test bytes
        p = urllib.parse.urlsplit(b"http://www.example.net:foo")
        self.assertEqual(p.netloc, b"www.example.net:foo")
        self.assertRaises(ValueError, lambda: p.port)

        p = urllib.parse.urlparse(b"http://www.example.net:foo")
        self.assertEqual(p.netloc, b"www.example.net:foo")
        self.assertRaises(ValueError, lambda: p.port)

    def test_attributes_without_netloc(self):
        # This example is straight from RFC 3261.  It looks like it
        # should allow the username, hostname, and port to be filled
        # in, but doesn't.  Since it's a URI and doesn't use the
        # scheme://netloc syntax, the netloc and related attributes
        # should be left empty.
        uri = "sip:alice@atlanta.com;maddr=239.255.255.1;ttl=15"
        p = urllib.parse.urlsplit(uri)
        self.assertEqual(p.netloc, "")
        self.assertEqual(p.username, None)
        self.assertEqual(p.password, None)
        self.assertEqual(p.hostname, None)
        self.assertEqual(p.port, None)
        self.assertEqual(p.geturl(), uri)

        p = urllib.parse.urlparse(uri)
        self.assertEqual(p.netloc, "")
        self.assertEqual(p.username, None)
        self.assertEqual(p.password, None)
        self.assertEqual(p.hostname, None)
        self.assertEqual(p.port, None)
        self.assertEqual(p.geturl(), uri)

        # You guessed it, repeating the test with bytes input
        uri = b"sip:alice@atlanta.com;maddr=239.255.255.1;ttl=15"
        p = urllib.parse.urlsplit(uri)
        self.assertEqual(p.netloc, b"")
        self.assertEqual(p.username, None)
        self.assertEqual(p.password, None)
        self.assertEqual(p.hostname, None)
        self.assertEqual(p.port, None)
        self.assertEqual(p.geturl(), uri)

        p = urllib.parse.urlparse(uri)
        self.assertEqual(p.netloc, b"")
        self.assertEqual(p.username, None)
        self.assertEqual(p.password, None)
        self.assertEqual(p.hostname, None)
        self.assertEqual(p.port, None)
        self.assertEqual(p.geturl(), uri)

    def test_noslash(self):
        # Issue 1637: http://foo.com?query is legal
        self.assertEqual(
            urllib.parse.urlparse("http://example.com?blahblah=/foo"),
            ("http", "example.com", "", "", "blahblah=/foo", ""),
        )
        self.assertEqual(
            urllib.parse.urlparse(b"http://example.com?blahblah=/foo"),
            (b"http", b"example.com", b"", b"", b"blahblah=/foo", b""),
        )

    def test_withoutscheme(self):
        # Test urlparse without scheme
        # Issue 754016: urlparse goes wrong with IP:port without scheme
        # RFC 1808 specifies that netloc should start with //, urlparse expects
        # the same, otherwise it classifies the portion of url as path.
        self.assertEqual(urllib.parse.urlparse("path"), ("", "", "path", "", "", ""))
        self.assertEqual(
            urllib.parse.urlparse("//www.python.org:80"), ("", "www.python.org:80", "", "", "", "")
        )
        self.assertEqual(
            urllib.parse.urlparse("http://www.python.org:80"),
            ("http", "www.python.org:80", "", "", "", ""),
        )
        # Repeat for bytes input
        self.assertEqual(urllib.parse.urlparse(b"path"), (b"", b"", b"path", b"", b"", b""))
        self.assertEqual(
            urllib.parse.urlparse(b"//www.python.org:80"),
            (b"", b"www.python.org:80", b"", b"", b"", b""),
        )
        self.assertEqual(
            urllib.parse.urlparse(b"http://www.python.org:80"),
            (b"http", b"www.python.org:80", b"", b"", b"", b""),
        )

    def test_portseparator(self):
        # Issue 754016 makes changes for port separator ':' from scheme separator
        self.assertEqual(urllib.parse.urlparse("path:80"), ("", "", "path:80", "", "", ""))
        self.assertEqual(urllib.parse.urlparse("http:"), ("http", "", "", "", "", ""))
        self.assertEqual(urllib.parse.urlparse("https:"), ("https", "", "", "", "", ""))
        self.assertEqual(
            urllib.parse.urlparse("http://www.python.org:80"),
            ("http", "www.python.org:80", "", "", "", ""),
        )
        # As usual, need to check bytes input as well
        self.assertEqual(urllib.parse.urlparse(b"path:80"), (b"", b"", b"path:80", b"", b"", b""))
        self.assertEqual(urllib.parse.urlparse(b"http:"), (b"http", b"", b"", b"", b"", b""))
        self.assertEqual(urllib.parse.urlparse(b"https:"), (b"https", b"", b"", b"", b"", b""))
        self.assertEqual(
            urllib.parse.urlparse(b"http://www.python.org:80"),
            (b"http", b"www.python.org:80", b"", b"", b"", b""),
        )

    def test_usingsys(self):
        # Issue 3314: sys module is used in the error
        self.assertRaises(TypeError, urllib.parse.urlencode, "foo")

    def test_anyscheme(self):
        # Issue 7904: s3://foo.com/stuff has netloc "foo.com".
        self.assertEqual(
            urllib.parse.urlparse("s3://foo.com/stuff"), ("s3", "foo.com", "/stuff", "", "", "")
        )
        self.assertEqual(
            urllib.parse.urlparse("x-newscheme://foo.com/stuff"),
            ("x-newscheme", "foo.com", "/stuff", "", "", ""),
        )
        self.assertEqual(
            urllib.parse.urlparse("x-newscheme://foo.com/stuff?query#fragment"),
            ("x-newscheme", "foo.com", "/stuff", "", "query", "fragment"),
        )
        self.assertEqual(
            urllib.parse.urlparse("x-newscheme://foo.com/stuff?query"),
            ("x-newscheme", "foo.com", "/stuff", "", "query", ""),
        )

        # And for bytes...
        self.assertEqual(
            urllib.parse.urlparse(b"s3://foo.com/stuff"),
            (b"s3", b"foo.com", b"/stuff", b"", b"", b""),
        )
        self.assertEqual(
            urllib.parse.urlparse(b"x-newscheme://foo.com/stuff"),
            (b"x-newscheme", b"foo.com", b"/stuff", b"", b"", b""),
        )
        self.assertEqual(
            urllib.parse.urlparse(b"x-newscheme://foo.com/stuff?query#fragment"),
            (b"x-newscheme", b"foo.com", b"/stuff", b"", b"query", b"fragment"),
        )
        self.assertEqual(
            urllib.parse.urlparse(b"x-newscheme://foo.com/stuff?query"),
            (b"x-newscheme", b"foo.com", b"/stuff", b"", b"query", b""),
        )

    def _test_mixed_types_rejected(self):
        # Several functions that process either strings or ASCII encoded bytes
        # accept multiple arguments. Check they reject mixed type input
        with self.assertRaisesRegex(TypeError, "Cannot mix str"):
            urllib.parse.urlparse("www.python.org", b"http")
        with self.assertRaisesRegex(TypeError, "Cannot mix str"):
            urllib.parse.urlparse(b"www.python.org", "http")
        with self.assertRaisesRegex(TypeError, "Cannot mix str"):
            urllib.parse.urlsplit("www.python.org", b"http")
        with self.assertRaisesRegex(TypeError, "Cannot mix str"):
            urllib.parse.urlsplit(b"www.python.org", "http")
        with self.assertRaisesRegex(TypeError, "Cannot mix str"):
            urllib.parse.urlunparse((b"http", "www.python.org", "", "", "", ""))
        with self.assertRaisesRegex(TypeError, "Cannot mix str"):
            urllib.parse.urlunparse(("http", b"www.python.org", "", "", "", ""))
        with self.assertRaisesRegex(TypeError, "Cannot mix str"):
            urllib.parse.urlunsplit((b"http", "www.python.org", "", "", ""))
        with self.assertRaisesRegex(TypeError, "Cannot mix str"):
            urllib.parse.urlunsplit(("http", b"www.python.org", "", "", ""))
        with self.assertRaisesRegex(TypeError, "Cannot mix str"):
            urllib.parse.urljoin("http://python.org", b"http://python.org")
        with self.assertRaisesRegex(TypeError, "Cannot mix str"):
            urllib.parse.urljoin(b"http://python.org", "http://python.org")

    def _check_result_type(self, str_type, num_args):
        #        num_args = len(str_type._fields)
        bytes_type = str_type._encoded_counterpart
        self.assertIs(bytes_type._decoded_counterpart, str_type)
        str_args = ("",) * num_args
        bytes_args = (b"",) * num_args
        str_result = str_type(*str_args)
        bytes_result = bytes_type(*bytes_args)
        encoding = "ascii"
        errors = "strict"
        self.assertEqual(str_result, str_args)
        self.assertEqual(bytes_result.decode(), str_args)
        self.assertEqual(bytes_result.decode(), str_result)
        self.assertEqual(bytes_result.decode(encoding), str_args)
        self.assertEqual(bytes_result.decode(encoding), str_result)
        self.assertEqual(bytes_result.decode(encoding, errors), str_args)
        self.assertEqual(bytes_result.decode(encoding, errors), str_result)
        self.assertEqual(bytes_result, bytes_args)
        self.assertEqual(str_result.encode(), bytes_args)
        self.assertEqual(str_result.encode(), bytes_result)
        self.assertEqual(str_result.encode(encoding), bytes_args)
        self.assertEqual(str_result.encode(encoding), bytes_result)
        self.assertEqual(str_result.encode(encoding, errors), bytes_args)
        self.assertEqual(str_result.encode(encoding, errors), bytes_result)

    def test_result_pairs(self):
        # Check encoding and decoding between result pairs
        result_types = [
            (urllib.parse.DefragResult, 2),
            (urllib.parse.SplitResult, 5),
            (urllib.parse.ParseResult, 6),
        ]
        for result_type in result_types:
            self._check_result_type(*result_type)

    def _test_parse_qs_encoding(self):
        result = urllib.parse.parse_qs("key=\u0141%E9", encoding="latin-1")
        self.assertEqual(result, {"key": ["\u0141\xE9"]})
        result = urllib.parse.parse_qs("key=\u0141%C3%A9", encoding="utf-8")
        self.assertEqual(result, {"key": ["\u0141\xE9"]})
        result = urllib.parse.parse_qs("key=\u0141%C3%A9", encoding="ascii")
        self.assertEqual(result, {"key": ["\u0141\ufffd\ufffd"]})
        result = urllib.parse.parse_qs("key=\u0141%E9-", encoding="ascii")
        self.assertEqual(result, {"key": ["\u0141\ufffd-"]})
        result = urllib.parse.parse_qs("key=\u0141%E9-", encoding="ascii", errors="ignore")
        self.assertEqual(result, {"key": ["\u0141-"]})

    def _test_parse_qsl_encoding(self):
        result = urllib.parse.parse_qsl("key=\u0141%E9", encoding="latin-1")
        self.assertEqual(result, [("key", "\u0141\xE9")])
        result = urllib.parse.parse_qsl("key=\u0141%C3%A9", encoding="utf-8")
        self.assertEqual(result, [("key", "\u0141\xE9")])
        result = urllib.parse.parse_qsl("key=\u0141%C3%A9", encoding="ascii")
        self.assertEqual(result, [("key", "\u0141\ufffd\ufffd")])
        result = urllib.parse.parse_qsl("key=\u0141%E9-", encoding="ascii")
        self.assertEqual(result, [("key", "\u0141\ufffd-")])
        result = urllib.parse.parse_qsl("key=\u0141%E9-", encoding="ascii", errors="ignore")
        self.assertEqual(result, [("key", "\u0141-")])

    def test_splitnport(self):
        # Normal cases are exercised by other tests; ensure that we also
        # catch cases with no port specified. (testcase ensuring coverage)
        result = urllib.parse.splitnport("parrot:88")
        self.assertEqual(result, ("parrot", 88))
        result = urllib.parse.splitnport("parrot")
        self.assertEqual(result, ("parrot", -1))
        result = urllib.parse.splitnport("parrot", 55)
        self.assertEqual(result, ("parrot", 55))
        result = urllib.parse.splitnport("parrot:")
        self.assertEqual(result, ("parrot", None))

    def test_splitquery(self):
        # Normal cases are exercised by other tests; ensure that we also
        # catch cases with no port specified (testcase ensuring coverage)
        result = urllib.parse.splitquery("http://python.org/fake?foo=bar")
        self.assertEqual(result, ("http://python.org/fake", "foo=bar"))
        result = urllib.parse.splitquery("http://python.org/fake?foo=bar?")
        self.assertEqual(result, ("http://python.org/fake?foo=bar", ""))
        result = urllib.parse.splitquery("http://python.org/fake")
        self.assertEqual(result, ("http://python.org/fake", None))

    def test_splitvalue(self):
        # Normal cases are exercised by other tests; test pathological cases
        # with no key/value pairs. (testcase ensuring coverage)
        result = urllib.parse.splitvalue("foo=bar")
        self.assertEqual(result, ("foo", "bar"))
        result = urllib.parse.splitvalue("foo=")
        self.assertEqual(result, ("foo", ""))
        result = urllib.parse.splitvalue("foobar")
        self.assertEqual(result, ("foobar", None))

    def test_to_bytes(self):
        result = urllib.parse.to_bytes("http://www.python.org")
        self.assertEqual(result, "http://www.python.org")

    #        self.assertRaises(UnicodeError, urllib.parse.to_bytes,
    #                          'http://www.python.org/medi\u00e6val')

    def test_urlencode_sequences(self):
        # Other tests incidentally urlencode things; test non-covered cases:
        # Sequence and object values.
        result = urllib.parse.urlencode({"a": [1, 2], "b": (3, 4, 5)}, True)
        # we cannot rely on ordering here
        assert set(result.split("&")) == {"a=1", "a=2", "b=3", "b=4", "b=5"}

        class Trivial:
            def __str__(self):
                return "trivial"

        result = urllib.parse.urlencode({"a": Trivial()}, True)
        self.assertEqual(result, "a=trivial")

    def test_quote_from_bytes(self):
        self.assertRaises(TypeError, urllib.parse.quote_from_bytes, "foo")
        result = urllib.parse.quote_from_bytes(b"archaeological arcana")
        self.assertEqual(result, "archaeological%20arcana")
        result = urllib.parse.quote_from_bytes(b"")
        self.assertEqual(result, "")

    def test_unquote_to_bytes(self):
        result = urllib.parse.unquote_to_bytes("abc%20def")
        self.assertEqual(result, b"abc def")
        result = urllib.parse.unquote_to_bytes("")
        self.assertEqual(result, b"")

    def test_quote_errors(self):
        self.assertRaises(TypeError, urllib.parse.quote, b"foo", encoding="utf-8")
        self.assertRaises(TypeError, urllib.parse.quote, b"foo", errors="strict")

    def test_issue14072(self):
        p1 = urllib.parse.urlsplit("tel:+31-641044153")
        self.assertEqual(p1.scheme, "tel")
        self.assertEqual(p1.path, "+31-641044153")
        p2 = urllib.parse.urlsplit("tel:+31641044153")
        self.assertEqual(p2.scheme, "tel")
        self.assertEqual(p2.path, "+31641044153")
        # assert the behavior for urlparse
        p1 = urllib.parse.urlparse("tel:+31-641044153")
        self.assertEqual(p1.scheme, "tel")
        self.assertEqual(p1.path, "+31-641044153")
        p2 = urllib.parse.urlparse("tel:+31641044153")
        self.assertEqual(p2.scheme, "tel")
        self.assertEqual(p2.path, "+31641044153")

    def test_telurl_params(self):
        p1 = urllib.parse.urlparse("tel:123-4;phone-context=+1-650-516")
        self.assertEqual(p1.scheme, "tel")
        self.assertEqual(p1.path, "123-4")
        self.assertEqual(p1.params, "phone-context=+1-650-516")

        p1 = urllib.parse.urlparse("tel:+1-201-555-0123")
        self.assertEqual(p1.scheme, "tel")
        self.assertEqual(p1.path, "+1-201-555-0123")
        self.assertEqual(p1.params, "")

        p1 = urllib.parse.urlparse("tel:7042;phone-context=example.com")
        self.assertEqual(p1.scheme, "tel")
        self.assertEqual(p1.path, "7042")
        self.assertEqual(p1.params, "phone-context=example.com")

        p1 = urllib.parse.urlparse("tel:863-1234;phone-context=+1-914-555")
        self.assertEqual(p1.scheme, "tel")
        self.assertEqual(p1.path, "863-1234")
        self.assertEqual(p1.params, "phone-context=+1-914-555")


def test_main():
    support.run_unittest(UrlParseTestCase)


if __name__ == "__main__":
    test_main()



import os
import time

n = os.uname().machine
if "EK-RA6M2" in n:
    i2c_id = 2
elif "EK-RA6M1" in n:
    i2c_id = 0
elif ("RA4M1 CLICKER" in n) or ("EK-RA4M1" in n) or ("EK-RA4W1" in n):
    print("SKIP")
    raise SystemExit
else:
    i2c_id = 0

from machine import I2C

i2c = I2C(i2c_id)



"""
categories: Core
description: f-strings cannot support expressions that require parsing to resolve unbalanced nested braces and brackets
cause: MicroPython is optimised for code space.
workaround: Always use balanced braces and brackets in expressions inside f-strings
"""

print(f'{"hello { world"}')
print(f'{"hello ] world"}')



# test "is" and "is not" with literal arguments
# these raise a SyntaxWarning in CPython because the results are
# implementation dependent; see https://bugs.python.org/issue34850

print(1 is 1)
print(1 is 2)
print(1 is not 1)
print(1 is not 2)

print("a" is "a")
print("a" is "b")
print("a" is not "a")
print("a" is not "b")
